#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

from earthkit.data import from_source
from earthkit.data.testing import earthkit_test_data_file


def test_netcdf_concat():
    ds1 = from_source("file", earthkit_test_data_file("era5_2t_1.nc"))
    ds2 = from_source("file", earthkit_test_data_file("era5_2t_2.nc"))
    ds = ds1 + ds2

    assert len(ds) == 2
    md = ds1.metadata("variable") + ds2.metadata("variable")
    assert ds.metadata("variable") == md

    assert ds[0].datetime() == {
        "base_time": datetime.datetime(2021, 3, 1, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 1, 12, 0),
    }
    assert ds[1].datetime() == {
        "base_time": datetime.datetime(2021, 3, 2, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 2, 12, 0),
    }
    assert ds.datetime() == {
        "base_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
        "valid_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
    }


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from abc import ABCMeta, abstractmethod

from earthkit.data.core.constants import DATETIME


class Metadata(metaclass=ABCMeta):
    r"""Base class to represent metadata.

    Metadata is a dict-like immutable object. Modification is possible
    with :obj:`override`, which always creates a new object.

    Implemented in subclasses: :obj:`RawMetadata`, :obj:`GribMetadata`.

    Examples
    --------
    - :ref:`/examples/metadata.ipynb`

    """

    def __iter__(self):
        """Return an iterator over the metadata keys."""
        return iter(self.keys())

    @abstractmethod
    def __len__(self):
        r"""Return the number of metadata entries."""

    def __getitem__(self, key):
        r"""Return the value for ``key``.

        Raises
        ------
        KeyError
            When ``key`` is not available.
        """
        return self._get_key(key)

    @abstractmethod
    def __contains__(self, key):
        r"""Check if ``key`` is available.

        Returns
        -------
        bool
        """
        pass

    @abstractmethod
    def keys(self):
        r"""Return a new view of the metadata keys.

        Returns
        -------
        Iterable of str

        """
        pass

    @abstractmethod
    def items(self):
        r"""Return a new view of the metadata items.

        Returns
        -------
        Iterable of :obj:`(key,value)` pairs

        """
        pass

    def get(self, key, default=None):
        r"""Return the value for ``key``.

        Parameters
        ----------
        key: str
            Metadata key
        default:
            When ``key`` is not found ``default`` is returned.
        """
        return self._get_key(key, default=default, raise_on_missing=False)

    def _get_key(self, key, astype=None, default=None, raise_on_missing=True):
        r"""Return the value for ``key``.

        Parameters
        ----------
        key: str
            Metadata key
        default: value
            Specify the default value for ``key``. Returned when ``key``
            is not found or its value is a missing value and raise_on_missing is ``False``.
        raise_on_missing: bool
            When it is True raises an exception if ``key`` is not found or
            it has a missing value.

        Returns
        -------
        value
            Returns the ``key`` value. Returns ``default`` if ``key`` is not found
            or it has a missing value and ``raise_on_missing`` is False.

        Raises
        ------
        KeyError
            If ``raise_on_missing`` is True and ``key`` is not found or it has
            a missing value.

        """
        # check for custom keys
        is_custom, v = self._get_custom_key(
            key, default=default, raise_on_missing=raise_on_missing
        )
        if is_custom:
            return v

        return self._get_internal_key(
            key, astype=astype, default=default, raise_on_missing=raise_on_missing
        )

    @abstractmethod
    def _get_internal_key(self, key, astype=None, default=None, raise_on_missing=True):
        pass

    def _get_custom_key(self, key, default=None, raise_on_missing=True):
        try:
            if key == DATETIME:
                return True, self.datetime().get("valid_time")
        except Exception as e:
            if not raise_on_missing:
                return default
            else:
                raise KeyError(f"{key}, reason={e}")
        return False, None

    @abstractmethod
    def override(self, *args, **kwargs):
        r"""Change the metadata values and return a new object.

        :obj:`override` accepts another :obj:`Metadata` or a dict or
        an iterable of key/value pairs (as tuples or other iterables of length two).
        If keyword arguments are specified, the metadata is then updated with those
        key/value pairs.

        Examples
        --------
        >>> other = RawMetadata({"key1": 1, "key2": 2})
        >>> m1 = m.override(other)
        >>> m1 = m.override({"key1": 1, "key2": 2})
        >>> m1 = m.override([("key1", 1), ("key2", 2)])
        >>> m1 = m.override(key1=1, key2=2)
        """
        pass

    def namespaces(self):
        r"""Return the available namespaces.

        Returns
        -------
        list of str
        """
        return []

    def as_namespace(self, namespace=None):
        r"""Return all the keys/values from a namespace.

        Parameters
        ----------
        namespace: str, None
            The namespace. When `namespace` is None or an empty :obj:`str` all the available
            keys/values are returned.

        Returns
        -------
        dict
            All the keys/values from the `namespace`.

        """
        if namespace is None or namespace == "":
            return {k: v for k, v in self.items()}
        return {}

    def dump(self, **kwargs):
        r"""Generate a dump from the metadata content."""
        return None

    def datetime(self):
        r"""Return the date and time.

        Returns
        -------
        dict of datatime.datetime or None
            Dict with items "base_time" and "valid_time". None if
            date and time are not available.
        """
        return None

    @property
    def geography(self):
        r""":obj:`Geography`: Get geography description.

        If it is not available None is returned.
        """
        return None

    def ls_keys(self):
        r"""Return the keys to be used with the :meth:`ls` method."""
        return []

    def describe_keys(self):
        r"""Return the keys to be used with the :meth:`describe` method."""
        return []

    def index_keys(self):
        r"""Return the keys to be used with the :meth:`indices` method."""
        return []


class RawMetadata(Metadata):
    r"""Metadata implementation based on key/value pairs.

    >>> from earthkit.data.core.metadata import RawMetadata
    >>> md = RawMetadata({"shortName": "2t", "perturbationNumber": 5})
    >>> md = RawMetadata([("shortName", "2t"), ("perturbationNumber", 5)])
    >>> md = RawMetadata(shortName="2t", perturbationNumber=5)
    >>> md
    RawMetadata({'shortName': '2t', 'perturbationNumber': 5})
    >>> md2 = RawMetadata(md)
    >>> md2
    RawMetadata({'shortName': '2t', 'perturbationNumber': 5})

    Value access:

    >>> "shortName" in md
    True
    >>> md["shortName"]
    '2t'
    >>> md.get("shortName")
    '2t'
    >>> "step" in md
    False
    >>> md.get("step")
    >>> md.get("step", 0)
    0

    Examples
    --------
    - :ref:`/examples/metadata.ipynb`

    """

    def __init__(self, *args, **kwargs):
        self._d = dict(*args, **kwargs)

    def __len__(self):
        return len(self._d)

    def keys(self):
        return self._d.keys()

    def items(self):
        return self._d.items()

    def __contains__(self, key):
        return key in self._d

    def _get_internal_key(self, key, astype=None, default=None, raise_on_missing=True):
        if not raise_on_missing:
            v = self._d.get(key, default)
        else:
            v = self._d[key]

        if astype is not None:
            try:
                return astype(v)
            except Exception:
                return None
        return v

    def override(self, *args, **kwargs):
        d = dict(**self._d)
        d.update(*args, **kwargs)
        return RawMetadata(d)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._d.__repr__()})"



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import pytest

from earthkit.data import from_source
from earthkit.data.testing import NO_FDB


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_FDB, reason="No access to FDB")
def test_fdb_grib_stream():
    d = datetime.datetime.now() - datetime.timedelta(days=2)
    request = {
        "class": "od",
        "expver": "0001",
        "stream": "oper",
        "date": d.strftime("%Y%m%d"),
        "time": [0, 12],
        "domain": "g",
        "type": "an",
        "levtype": "sfc",
        "step": 0,
        "param": [151, 167],
    }

    ds = from_source("fdb", request)
    cnt = sum([1 for f in ds])
    assert cnt == 4


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_FDB, reason="No access to FDB")
def test_fdb_grib_file():
    d = datetime.datetime.now() - datetime.timedelta(days=2)
    request = {
        "class": "od",
        "expver": "0001",
        "stream": "oper",
        "date": d.strftime("%Y%m%d"),
        "time": [0, 12],
        "domain": "g",
        "type": "an",
        "levtype": "sfc",
        "step": 0,
        "param": [151, 167],
    }

    ds = from_source("fdb", request, stream=False)
    assert len(ds) == 4


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data.decorators import alias_argument, normalize


def func_x(x):
    return x


def func_xy(x, y):
    return x, y


def func_xyz(x, y, z):
    return x, y, z


def test_alias_argument_1():
    f = alias_argument()(func_x)
    assert f(1) == 1
    assert f(x=1) == 1
    with pytest.raises(TypeError):
        assert f(unk=1) == 1

    f = alias_argument(x=["alias"])(func_x)
    assert f(alias=1) == 1
    assert f(x=1) == 1

    f = alias_argument(x="alias")(func_x)
    assert f(alias=1) == 1
    assert f(x=1) == 1

    with pytest.raises(ValueError):
        f = alias_argument(x={"alias": "value"})(func_x)

    f = alias_argument(x=["alias", "blias"])(func_x)
    assert f(alias=1) == 1

    with pytest.raises(TypeError):
        f = alias_argument(unk=["x"])(func_x)
        assert f(x=1) == 1


def test_alias_argument_1_bis():
    f = alias_argument("x", ["alias"])(func_x)
    assert f(alias=1) == 1

    f = alias_argument("x", "alias")(func_x)
    assert f(alias=1) == 1


def test_alias_argument_2_bis():
    f = alias_argument("x", ["alias_x"])(func_xy)
    f = alias_argument("y", ["alias_y"])(f)
    assert f(alias_x=1, alias_y=2) == (1, 2)


def test_alias_argument_2():
    f = alias_argument(y=["alias", "blias"])(func_xy)
    assert f(1, alias=2) == (1, 2)

    f = alias_argument(y=["alias", "blias"])(func_xy)
    assert f(1, alias=2) == (1, 2)
    assert f(1, blias=2) == (1, 2)
    assert f(1, y=2) == (1, 2)
    assert f(x=1, alias=2) == (1, 2)
    assert f(x=1, blias=2) == (1, 2)
    assert f(x=1, y=2) == (1, 2)


def test_alias_argument_3():
    f = alias_argument(y=["alias", "blias"])(func_xyz)
    assert f(1, alias=2, z=3) == (1, 2, 3)
    assert f(1, blias=2, z=3) == (1, 2, 3)
    assert f(z=3, alias=2, x=1) == (1, 2, 3)
    assert f(x=1, alias=2, z=3) == (1, 2, 3)
    assert f(x=1, z=3, blias=2) == (1, 2, 3)
    assert f(x=1, y=2, z=3) == (1, 2, 3)

    f = alias_argument(y=["y_alias"], x=["x_alias"])(func_xyz)
    assert f(y_alias=2, x_alias=1, z=3) == (1, 2, 3)

    with pytest.raises(ValueError):
        f = alias_argument(x=["alias"], y=["alias"])(func_xyz)
    with pytest.raises(ValueError):
        f = alias_argument(x=["alias", "blias"], y=["alias"])(func_xyz)
    with pytest.raises(ValueError):
        f = alias_argument(x=["alias", "blias"], y=["alias", "blias"])(func_xyz)


@pytest.mark.parametrize("typ", [str, int, float])
def test_aliases_grib_paramid_mutiple_false(typ):
    _131 = typ(131)
    aliases_grib_paramid = normalize(
        "x",
        type=typ,
        aliases={"u": typ(131), "v": typ(132)},
        multiple=False,
    )(func_x)
    assert aliases_grib_paramid("u") == _131
    assert aliases_grib_paramid(131) == _131
    assert aliases_grib_paramid("131") == _131

    # one-element list/tuple
    assert aliases_grib_paramid(("131",)) == _131
    assert aliases_grib_paramid(["131"]) == _131

    # list/tuple
    with pytest.raises(TypeError):
        aliases_grib_paramid(["131", "v"])

    # empty list/tuple
    with pytest.raises(TypeError):
        aliases_grib_paramid([])
    with pytest.raises(TypeError):
        aliases_grib_paramid(tuple([]))


@pytest.mark.parametrize(
    "typ,_131,_132", [(str, "131", "132"), (int, 131, 132), (float, 131.0, 132.0)]
)
def test_aliases_grib_paramid_mutiple_true(typ, _131, _132):
    aliases_grib_paramid = normalize(
        "x",
        type=typ,
        aliases={"u": typ(131), "v": typ(132)},
        multiple=True,
    )(func_x)
    # single values
    assert aliases_grib_paramid("u") == [_131]
    assert aliases_grib_paramid(131) == [_131]
    assert aliases_grib_paramid("131") == [_131]

    # one-element list/tuple
    assert aliases_grib_paramid(("131",)) == [_131]
    assert aliases_grib_paramid(["131"]) == [_131]

    # list/tuple
    assert aliases_grib_paramid(["131", "v"]) == [_131, _132]
    assert aliases_grib_paramid([131, "v"]) == [_131, _132]
    assert aliases_grib_paramid(["u", "v"]) == [_131, _132]
    assert aliases_grib_paramid(("u", "v")) == [_131, _132]
    assert aliases_grib_paramid([]) == []
    assert aliases_grib_paramid(tuple([])) == []


@pytest.mark.parametrize(
    "typ,_131,_132", [(str, "131", "132"), (int, 131, 132), (float, 131.0, 132.0)]
)
def test_aliases_mutiple_none(typ, _131, _132):
    aliases_func = normalize(
        "x",
        type=typ,
        aliases={"u": _131, "v": _132},
    )(func_x)
    # single values
    assert aliases_func("u") == _131
    assert aliases_func(131) == _131
    assert aliases_func("131") == _131

    # one-element list/tuple
    assert aliases_func(("131",)) == (_131,)
    assert aliases_func(["131"]) == [_131]

    # list/tuple
    assert aliases_func(["131", "v"]) == [_131, _132]
    assert aliases_func([131, "v"]) == [_131, _132]
    assert aliases_func(["u", "v"]) == [_131, _132]
    assert aliases_func(("u", "v")) == (_131, _132)
    assert aliases_func([]) == []
    assert aliases_func(tuple([])) == ()


def test_aliases():
    @normalize("x", aliases={"one": 1})
    def f(x):
        return x

    assert f(1) == 1
    assert f("one") == 1


if __name__ == "__main__":
    # import logging
    # logging.basicConfig(level=logging.WARNING)
    # test_aliases_grib_paramid_mutiple_true()
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from abc import abstractmethod
from collections import defaultdict

from earthkit.data.core import Base
from earthkit.data.core.index import Index
from earthkit.data.decorators import cached_method
from earthkit.data.utils.metadata import metadata_argument


class Field(Base):
    r"""Represents a Field."""

    def __init__(self, metadata=None):
        self.__metadata = metadata

    @property
    @abstractmethod
    def values(self):
        r"""ndarray: Get the values stored in the field as a 1D ndarray."""
        self._not_implemented()

    def _make_metadata(self):
        r"""Create a field metadata object."""
        self._not_implemented()

    @property
    def _metadata(self):
        r"""Metadata: Get the object representing the field's metadata."""
        if self.__metadata is None:
            self.__metadata = self._make_metadata()
        return self.__metadata

    def to_numpy(self, flatten=False, dtype=None):
        r"""Return the values stored in the field as an ndarray.

        Parameters
        ----------
        flatten: bool
            When it is True a flat ndarray is returned. Otherwise an ndarray with the field's
            :obj:`shape` is returned.
        dtype: str or dtype
            Typecode or data-type to which the array is cast.

        Returns
        -------
        ndarray
            Field values

        """
        values = self.values
        if not flatten:
            values = self.values.reshape(self.shape)
        if dtype is not None:
            values = values.astype(dtype)
        return values

    def data(self, keys=("lat", "lon", "value"), flatten=False):
        r"""Return the values and/or the geographical coordinates for each grid point.

        Parameters
        ----------
        keys: :obj:`str`, :obj:`list` or :obj:`tuple`
            Specifies the type of data to be returned. Any combination of "lat", "lon" and "value"
            is allowed here.
        flatten: bool
            When it is True a flat ndarray per key is returned. Otherwise an ndarray with the field's
            :obj:`shape` is returned for each key.

        Returns
        -------
        ndarray
            An ndarray containing one ndarray per key is returned
            (following the order in ``keys``). When ``keys`` is a single value only the
            ndarray belonging to the key is returned.


        Examples
        --------
        - :ref:`/examples/grib_lat_lon_value.ipynb`

        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test6.grib")
        >>> d = ds[0].data()
        >>> d.shape
        (3, 7, 12)
        >>> d[0, 0, 0]  # first latitude
        90.0
        >>> d[1, 0, 0]  # first longitude
        0.0
        >>> d[2, 0, 0]  # first value
        272.56417847
        >>> d = ds[0].data(keys="lon")
        >>> d.shape
        (7, 12)
        >>> d[0, 0]  # first longitude
        0.0

        See Also
        --------
        to_latlon
        to_points
        to_numpy
        values

        """
        _keys = dict(
            lat=self._metadata.geography.latitudes,
            lon=self._metadata.geography.longitudes,
            value=lambda: self.values,
        )

        if isinstance(keys, str):
            keys = [keys]

        for k in keys:
            if k not in _keys:
                raise ValueError(f"data: invalid argument: {k}")

        r = [_keys[k]() for k in keys]
        if not flatten:
            shape = self.shape
            r = [x.reshape(shape) for x in r]
        if len(r) == 1:
            return r[0]
        else:
            import numpy as np

            return np.array(r)

    def to_points(self, flatten=False):
        r"""Return the geographical coordinates in the data's original
        Coordinate Reference System (CRS).

        Parameters
        ----------
        flatten: bool
            When it is True 1D ndarrays are returned. Otherwise ndarrays with the field's
            :obj:`shape` are returned.

        Returns
        -------
        dict
            Dictionary with items "x" and "y", containing the ndarrays of the x and
            y coordinates, respectively.

        Raises
        ------
        ValueError
            When the coordinates in the data's original CRS are not available.

        See Also
        --------
        to_latlon

        """
        x = self._metadata.geography.x()
        y = self._metadata.geography.y()
        if x is not None and y is not None:
            if not flatten:
                shape = self.shape
                x = x.reshape(shape)
                y = y.reshape(shape)
            return dict(x=x, y=y)
        elif self.projection().CARTOPY_CRS == "PlateCarree":
            lon, lat = self.data(("lon", "lat"), flatten=flatten)
            return dict(x=lon, y=lat)
        else:
            raise ValueError(
                "to_points(): geographical coordinates in original CRS are not available"
            )

    def to_latlon(self, flatten=False):
        r"""Return the latitudes/longitudes of all the gridpoints in the field.

        Parameters
        ----------
        flatten: bool
            When it is True 1D ndarrays are returned. Otherwise ndarrays with the field's
            :obj:`shape` are returned.

        Returns
        -------
        dict
            Dictionary with items "lat" and "lon", containing the ndarrays of the latitudes and
            longitudes, respectively.

        See Also
        --------
        to_points

        """
        lon, lat = self.data(("lon", "lat"), flatten=flatten)
        return dict(lat=lat, lon=lon)

    @property
    def shape(self):
        r"""tuple: Get the shape of the field.

        For structured grids the shape is a tuple in the form of (Nj, Ni) where:

        - ni: the number of gridpoints in i direction (longitude for a regular latitude-longitude grid)
        - nj: the number of gridpoints in j direction (latitude for a regular latitude-longitude grid)

        For other grid types the number of gridpoints is returned as ``(num,)``
        """
        return self._metadata.geography.shape()

    def projection(self):
        r"""Return information about the projection.

        Returns
        -------
        :obj:`Projection`

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> ds.projection()
        <Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to ...>
        Name: unknown
        Axis Info [cartesian]:
        - E[east]: Easting (unknown)
        - N[north]: Northing (unknown)
        - h[up]: Ellipsoidal height (metre)
        Area of Use:
        - undefined
        Coordinate Operation:
        - name: unknown
        - method: Equidistant Cylindrical
        Datum: Unknown based on WGS 84 ellipsoid
        - Ellipsoid: WGS 84
        - Prime Meridian: Greenwich
        >>> ds.projection().to_proj_string()
        '+proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to_meter=111319.4907932736 +no_defs +type=crs'
        """
        return self._metadata.geography.projection()

    def bounding_box(self):
        r"""Return the bounding box of the field.

        Returns
        -------
        :obj:`BoundingBox <data.utils.bbox.BoundingBox>`
        """
        return self._metadata.geography.bounding_box()

    def datetime(self):
        r"""Return the date and time of the field.

        Returns
        -------
        dict of datatime.datetime
            Dict with items "base_time" and "valid_time".


        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "tests/data/t_time_series.grib")
        >>> ds[4].datetime()
        {'base_time': datetime.datetime(2020, 12, 21, 12, 0),
        'valid_time': datetime.datetime(2020, 12, 21, 18, 0)}

        """
        return self._metadata.datetime()

    def metadata(self, *keys, astype=None, **kwargs):
        r"""Return metadata values from the field.

        When called without any arguments returns a :obj:`Metadata` object.

        Parameters
        ----------
        *keys: tuple
            Positional arguments specifying metadata keys. Can be empty, in this case all
            the keys from the specified ``namespace`` will
            be used. (See examples below).
        astype: type name, :obj:`list` or :obj:`tuple`
            Return types for ``keys``. A single value is accepted and applied to all the ``keys``.
            Otherwise, must have same the number of elements as ``keys``. Only used when
            ``keys`` is not empty.
        **kwargs: dict, optional
            Other keyword arguments:

            * namespace: :obj:`str`, :obj:`list`, :obj:`tuple`, :obj:`None` or :obj:`all`
                The namespace to choose the ``keys`` from. When ``keys`` is empty and ``namespace`` is
                :obj:`all` all the available namespaces will be used. When ``keys`` is non empty
                ``namespace`` cannot specify multiple values and it cannot be :obj:`all`. When
                ``namespace`` is None or empty str all the available keys will be used
                (without a namespace qualifier).

            * default: value, optional
                Specifies the same default value for all the ``keys`` specified. When ``default`` is
                **not present** and a key is not found or its value is a missing value
                :obj:`metadata` will raise KeyError.

        Returns
        -------
        single value, :obj:`list`, :obj:`tuple`, :obj:`dict` or :obj:`Metadata`
            - when called without any arguments returns a :obj:`Metadata` object
            - when ``keys`` is not empty:
                - returns single value when ``keys`` is a str
                - otherwise returns the same type as that of ``keys`` (:obj:`list` or :obj:`tuple`)
            - when ``keys`` is empty:
                - when ``namespace`` is None or an empty str returns a :obj:`dict` with all
                  the available keys and values
                - when ``namespace`` is :obj:`str` returns a :obj:`dict` with the keys and values
                  in that namespace
                - otherwise returns a :obj:`dict` with one item per namespace (dict of dict)

        Raises
        ------
        KeyError
            If no ``default`` is set and a key is not found in the message or it has a missing value.


        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")

        Calling without arguments:

        >>> r = ds[0].metadata()
        >>> r
        <earthkit.data.readers.grib.metadata.GribMetadata object at 0x164ace170>
        >>> r["name"]
        '2 metre temperature'

        Getting keys with their native type:

        >>> ds[0].metadata("param")
        '2t'
        >>> ds[0].metadata("param", "units")
        ('2t', 'K')
        >>> ds[0].metadata(("param", "units"))
        ('2t', 'K')
        >>> ds[0].metadata(["param", "units"])
        ['2t', 'K']
        >>> ds[0].metadata(["param"])
        ['2t']
        >>> ds[0].metadata("badkey")
        KeyError: 'badkey'
        >>> ds[0].metadata("badkey", default=None)
        <BLANKLINE>

        Prescribing key types:

        >>> ds[0].metadata("centre", astype=int)
        98
        >>> ds[0].metadata(["paramId", "centre"], astype=int)
        [167, 98]
        >>> ds[0].metadata(["centre", "centre"], astype=[int, str])
        [98, 'ecmf']

        Using namespaces:

        >>> ds[0].metadata(namespace="parameter")
        {'centre': 'ecmf', 'paramId': 167, 'units': 'K', 'name': '2 metre temperature', 'shortName': '2t'}
        >>> ds[0].metadata(namespace=["parameter", "vertical"])
        {'parameter': {'centre': 'ecmf', 'paramId': 167, 'units': 'K', 'name': '2 metre temperature',
         'shortName': '2t'},
         'vertical': {'typeOfLevel': 'surface', 'level': 0}}
        >>> r = ds[0].metadata(namespace=all)
        >>> r.keys()
        dict_keys(['default', 'ls', 'geography', 'mars', 'parameter', 'statistics', 'time', 'vertical'])
        >>> r = ds[0].metadata(namespace=None)
        >>> len(r)
        186
        >>> r["name"]
        '2 metre temperature'
        """
        # when called without arguments returns the metadata object
        if len(keys) == 0 and astype is None and len(kwargs) == 0:
            return self._metadata

        namespace = kwargs.pop("namespace", None)
        key, namespace, astype, key_arg_type = metadata_argument(
            *keys, namespace=namespace, astype=astype
        )

        assert isinstance(key, list)
        assert isinstance(namespace, (list, tuple))

        if key:
            assert isinstance(astype, (list, tuple))
            if namespace and namespace[0] != "default":
                key = [namespace[0] + "." + k for k in key]

            raise_on_missing = "default" not in kwargs
            default = kwargs.pop("default", None)

            r = [
                self._metadata._get_key(
                    k,
                    astype=kt,
                    default=default,
                    raise_on_missing=raise_on_missing,
                    **kwargs,
                )
                for k, kt in zip(key, astype)
            ]

            if key_arg_type == str:
                return r[0]
            elif key_arg_type == tuple:
                return tuple(r)
            else:
                return r
        elif namespace:
            if all in namespace:
                namespace = self._metadata.namespaces()

            r = {ns: self._metadata.as_namespace(ns) for ns in namespace}
            if len(r) == 1:
                return r[namespace[0]]
            else:
                return r
        else:
            return self._metadata.as_namespace(None)

    def dump(self, namespace=all, **kwargs):
        r"""Generate dump with all the metadata keys belonging to ``namespace``.

        In a Jupyter notebook it is represented as a tabbed interface.

        Parameters
        ----------
        namespace: :obj:`str`, :obj:`list`, :obj:`tuple`, :obj:`None` or :obj:`all`
            The namespace to dump. The following `namespace` values
            have a special meaning:

            - :obj:`all`: all the available namespaces will be used.
            - None or empty str: all the available keys will be used
                (without a namespace qualifier)

        **kwargs: dict, optional
            Other keyword arguments used for testing only

        Returns
        -------
        NamespaceDump
            Dict-like object with one item per namespace. In a Jupyter notebook represented
            as a tabbed interface to browse the dump contents.

        Examples
        --------
        :ref:`/examples/grib_metadata.ipynb`

        """
        return self._metadata.dump(namespace=namespace, **kwargs)

    def __getitem__(self, key):
        """Return the value of the metadata ``key``."""
        return self._metadata.get(key)

    @abstractmethod
    def message(self):
        r"""Return a buffer containing the encoded message for message based formats (e.g. GRIB).

        Returns
        -------
        bytes
        """
        self._not_implemented()

    def __repr__(self):
        return "%s(%s,%s,%s,%s,%s,%s)" % (
            self.__class__.__name__,
            self._metadata.get("shortName", None),
            self._metadata.get("levelist", None),
            self._metadata.get("date", None),
            self._metadata.get("time", None),
            self._metadata.get("step", None),
            self._metadata.get("number", None),
        )

    @abstractmethod
    def _attributes(self, names):
        result = {}
        for name in names:
            result[name] = self._metadata.get(name, None)
        return result


class FieldList(Index):
    r"""Represents a list of :obj:`Field` \s."""

    _indices = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def from_numpy(array, metadata):
        from earthkit.data.sources.numpy_list import NumpyFieldList

        return NumpyFieldList(array, metadata)

    @cached_method
    def _default_index_keys(self):
        if len(self) > 0:
            return self[0].metadata().index_keys()
        else:
            return []

    def _find_index_values(self, key):
        values = set()
        for f in self:
            v = f.metadata(key, default=None)
            if v is not None:
                values.add(v)
        return sorted(list(values))

    def _find_all_index_dict(self):
        indices = defaultdict(set)
        for f in self:
            for k in self._default_index_keys():
                v = f.metadata(k, default=None)
                if v is None:
                    continue
                indices[k].add(v)

        return {k: sorted(list(v)) for k, v in indices.items()}

    def indices(self, squeeze=False):
        r"""Return the unique, sorted values for a set of metadata keys (see below)
        from all the fields. Individual keys can be also queried by :obj:`index`.

        Parameters
        ----------
        squeeze : False
            When True only returns the metadata keys that have more than one values.

        Returns
        -------
        dict
            Unique, sorted metadata values from all the
            :obj:`GribField <data.readers.grib.codes.GribField>`\ s.

        See Also
        --------
        index

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/tuv_pl.grib")
        >>> ds.indices()
        {'class': ['od'], 'stream': ['oper'], 'levtype': ['pl'], 'type': ['an'],
        'expver': ['0001'], 'date': [20180801], 'time': [1200], 'domain': ['g'],
        'number': [0], 'levelist': [300, 400, 500, 700, 850, 1000],
        'param': ['t', 'u', 'v']}
        >>> ds.indices(squeeze=True)
        {'levelist': [300, 400, 500, 700, 850, 1000], 'param': ['t', 'u', 'v']}
        >>> ds.index("level")
        [300, 400, 500, 700, 850, 1000]

        By default :obj:`indices` uses the keys from the
        "mars" :xref:`eccodes_namespace`. Keys with no valid values are not included. Keys
        that :obj:`index` is called with are automatically added to the original set of keys
        used in :obj:`indices`.

        """
        if not self._indices:
            self._indices = self._find_all_index_dict()
        if squeeze:
            return {k: v for k, v in self._indices.items() if len(v) > 1}
        else:
            return self._indices

    def index(self, key):
        r"""Return the unique, sorted values of the specified metadata ``key`` from all the fields.
        ``key`` will be automatically added to the keys returned by :obj:`indices`.

        Parameters
        ----------
        key : str
            Metadata key.

        Returns
        -------
        list
            Unique, sorted values of ``key`` from all the
            :obj:`GribField <data.readers.grib.codes.GribField>`\ s.

        See Also
        --------
        index

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/tuv_pl.grib")
        >>> ds.index("level")
        [300, 400, 500, 700, 850, 1000]

        """
        if key in self.indices():
            return self.indices()[key]

        self._indices[key] = self._find_index_values(key)
        return self._indices[key]

    def to_numpy(self, **kwargs):
        r"""Return the field values as an ndarray. It is formed as the array of the
        :obj:`data.core.fieldlist.Field.to_numpy` values per field.

        Parameters
        ----------
        **kwargs: dict, optional
            Keyword arguments passed to :obj:`data.core.fieldlist.Field.to_numpy`

        Returns
        -------
        ndarray
            Array containing the field values.

        See Also
        --------
        values
        """
        import numpy as np

        return np.array([f.to_numpy(**kwargs) for f in self])

    @property
    def values(self):
        r"""ndarray: Get the field values as a 2D ndarray. It is formed as the array of
        :obj:`GribField.values <data.readers.grib.codes.GribField.values>` per field.

        See Also
        --------
        to_numpy


        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> for f in ds:
        ...     print(f.values.shape)
        ...
        (209,)
        (209,)
        >>> v = ds.values
        >>> v.shape
        (2, 209)
        >>> v[0][:3]
        array([262.78027344, 267.44726562, 268.61230469])

        """
        import numpy as np

        return np.array([f.values for f in self])

    def data(self, keys=("lat", "lon", "value"), flatten=False):
        r"""Return the values and/or the geographical coordinates.

        Only works when all the fields have the same grid geometry.

        Parameters
        ----------
        keys: :obj:`str`, :obj:`list` or :obj:`tuple`
            Specifies the type of data to be returned. Any combination of "lat", "lon" and "value"
            is allowed here.
        flatten: bool
            When it is True the "lat", "lon" arrays and the "value" arrays per field
            will all be flattened. Otherwise they will preserve the field's :obj:`shape`.

        Returns
        -------
        ndarray
            The elements of the ndarray (in the order of the ``keys``) are as follows:

            * the latitudes array from the first field  when "lat" is in ``keys``
            * the longitudes array from the first field when "lon" is in ``keys``
            * a values array per field when "values" is in ``keys``


        Raises
        ------
        ValueError
            When not all the fields have the same grid geometry.


        Examples
        --------
        - :ref:`/examples/grib_lat_lon_value.ipynb`

        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test6.grib")
        >>> len(ds)
        6
        >>> d = ds.data()
        >>> d.shape
        (8, 7, 12)
        >>> d[0, 0, 0]  # first latitude
        90.0
        >>> d[1, 0, 0]  # first longitude
        0.0
        >>> d[2:, 0, 0]  # first value per field
        array([272.56417847,  -6.28688049,   7.83348083, 272.53916931,
                -4.89837646,   8.66096497])
        >>> d = ds.data(keys="lon")
        >>> d.shape
        (1, 7, 12)
        >>> d[0, 0, 0]  # first longitude
        0.0

        See Also
        --------
        to_latlon
        to_points
        to_numpy
        values

        """
        import numpy as np

        if self._is_shared_grid():
            if isinstance(keys, str):
                keys = [keys]

            if "lat" in keys or "lon" in keys:
                latlon = self[0].to_latlon(flatten=flatten)

            r = []
            for k in keys:
                if k == "lat":
                    r.append(latlon["lat"])
                elif k == "lon":
                    r.append(latlon["lon"])
                elif k == "value":
                    r.extend([f.to_numpy(flatten=flatten) for f in self])
                else:
                    raise ValueError(f"data: invalid argument: {k}")

            return np.array(r)

        elif len(self) == 0:
            return np.array([])
        else:
            raise ValueError("Fields do not have the same grid geometry")

    def metadata(self, *args, **kwargs):
        r"""Return the metadata values for each field.

        Parameters
        ----------
        *args: tuple
            Positional arguments defining the metadata keys. Passed to
            :obj:`GribField.metadata() <data.readers.grib.codes.GribField.metadata>`
        **kwargs: dict, optional
            Keyword arguments passed to
            :obj:`GribField.metadata() <data.readers.grib.codes.GribField.metadata>`

        Returns
        -------
        list
            List with one item per :obj:`GribField <data.readers.grib.codes.GribField>`

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> ds.metadata("param")
        ['2t', 'msl']
        >>> ds.metadata("param", "units")
        [('2t', 'K'), ('msl', 'Pa')]
        >>> ds.metadata(["param", "units"])
        [['2t', 'K'], ['msl', 'Pa']]

        """
        result = []
        for s in self:
            result.append(s.metadata(*args, **kwargs))
        return result

    def ls(self, n=None, keys=None, extra_keys=None, namespace=None):
        r"""Generate a list like summary using a set of metadata keys.

        Parameters
        ----------
        n: int, None
            The number of :obj:`Field`\ s to be
            listed. None means all the messages, ``n > 0`` means fields from the front, while
            ``n < 0`` means fields from the back of the fieldlist.
        keys: list of str, dict, None
            Metadata keys. If it is None the following default set of keys will be used:  "centre",
            "shortName", "typeOfLevel", "level", "dataDate", "dataTime", "stepRange", "dataType",
            "number", "gridType". To specify a column title for each key in the output use a dict.
        extra_keys: list of str, dict, None
            List of additional keys to ``keys``. To specify a column title for each key in the output
            use a dict.
        namespace: str, None
            The namespace to choose the ``keys`` from. When it is set ``keys`` and
            ``extra_keys`` are omitted.

        Returns
        -------
        Pandas DataFrame
            DataFrame with one row per :obj:`Field`.

        See Also
        --------
        head
        tail

        """
        from earthkit.data.utils.summary import ls

        def _proc(keys, n):
            num = len(self)
            pos = slice(0, num)
            if n is not None:
                pos = slice(0, min(num, n)) if n > 0 else slice(num - min(num, -n), num)
            pos_range = range(pos.start, pos.stop)

            if "namespace" in keys:
                ns = keys.pop("namespace", None)
                for i in pos_range:
                    f = self[i]
                    v = f.metadata(namespace=ns)
                    if len(keys) > 0:
                        v.update(f._attributes(keys))
                    yield (v)
            else:
                for i in pos_range:
                    yield (self[i]._attributes(keys))

        _keys = (
            self._default_ls_keys() if namespace is None else dict(namespace=namespace)
        )
        return ls(_proc, _keys, n=n, keys=keys, extra_keys=extra_keys)

    @cached_method
    def _default_ls_keys(self):
        if len(self) > 0:
            return self[0].metadata().ls_keys()
        else:
            return []

    def head(self, n=5, **kwargs):
        r"""Generate a list like summary of the first ``n``
        :obj:`GribField <data.readers.grib.codes.GribField>`\ s using a set of metadata keys.
        Same as calling :obj:`ls` with ``n``.

        Parameters
        ----------
        n: int, None
            The number of messages (``n`` > 0) to be printed from the front.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`ls`.

        Returns
        -------
        Pandas DataFrame
            See  :obj:`ls`.

        See Also
        --------
        ls
        tail


        The following calls are equivalent:

            .. code-block:: python

                ds.head()
                ds.head(5)
                ds.head(n=5)
                ds.ls(5)
                ds.ls(n=5)

        """
        if n <= 0:
            raise ValueError("head: n must be > 0")
        return self.ls(n=n, **kwargs)

    def tail(self, n=5, **kwargs):
        r"""Generate a list like summary of the last ``n``
        :obj:`GribField <data.readers.grib.codes.GribField>`\ s using a set of metadata keys.
        Same as calling :obj:`ls` with ``-n``.

        Parameters
        ----------
        n: int, None
            The number of messages (``n`` > 0)  to be printed from the back.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`ls`.

        Returns
        -------
        Pandas DataFrame
            See  :obj:`ls`.

        See Also
        --------
        head
        ls


        The following calls are equivalent:

            .. code-block:: python

                ds.tail()
                ds.tail(5)
                ds.tail(n=5)
                ds.ls(-5)
                ds.ls(n=-5)

        """
        if n <= 0:
            raise ValueError("n must be > 0")
        return self.ls(n=-n, **kwargs)

    def describe(self, *args, **kwargs):
        r"""Generate a summary of the fieldlist."""
        from earthkit.data.utils.summary import format_describe

        def _proc():
            for f in self:
                yield (f._attributes(self._describe_keys()))

        return format_describe(_proc(), *args, **kwargs)

    @cached_method
    def _describe_keys(self):
        if len(self) > 0:
            return self[0].metadata().describe_keys()
        else:
            return []

    def datetime(self):
        r"""Return the unique, sorted list of dates and times in the fieldlist.

        Returns
        -------
        dict of datatime.datetime
            Dict with items "base_time" and "valid_time".


        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "tests/data/t_time_series.grib")
        >>> ds.datetime()
        {'base_time': [datetime.datetime(2020, 12, 21, 12, 0)],
        'valid_time': [
            datetime.datetime(2020, 12, 21, 12, 0),
            datetime.datetime(2020, 12, 21, 15, 0),
            datetime.datetime(2020, 12, 21, 18, 0),
            datetime.datetime(2020, 12, 21, 21, 0),
            datetime.datetime(2020, 12, 23, 12, 0)]}

        """
        base = set()
        valid = set()
        for s in self:
            d = s.datetime()
            base.add(d["base_time"])
            valid.add(d["valid_time"])
        return {"base_time": sorted(base), "valid_time": sorted(valid)}

    def to_points(self, **kwargs):
        r"""Return the geographical coordinates shared by all the fields in
        the data's original Coordinate Reference System (CRS).

        Parameters
        ----------
        **kwargs: dict, optional
            Keyword arguments passed to
            :obj:`GribField.to_points() <data.readers.grib.codes.GribField.to_points>`

        Returns
        -------
        dict
            Dictionary with items "x" and "y", containing the ndarrays of the x and
            y coordinates, respectively.

        Raises
        ------
        ValueError
            When not all the fields have the same grid geometry.
        """
        if self._is_shared_grid():
            return self[0].to_points(**kwargs)
        elif len(self) == 0:
            return dict(x=None, y=None)
        else:
            raise ValueError("Fields do not have the same grid geometry")

    def to_latlon(self, **kwargs):
        r"""Return the latitudes/longitudes shared by all the fields.

        Parameters
        ----------
        **kwargs: dict, optional
            Keyword arguments passed to
            :obj:`GribField.to_latlon() <data.readers.grib.codes.GribField.to_latlon>`

        Returns
        -------
        dict
            Dictionary with items "lat" and "lon", containing the ndarrays of the latitudes and
            longitudes, respectively.

        Raises
        ------
        ValueError
            When not all the fields have the same grid geometry

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> for f in ds:
        ...     print(f.shape)
        ...
        (11, 19)
        (11, 19)
        >>> r = ds.to_latlon()
        >>> for k, v in r.items():
        ...     print(f"{k}: shape={v.shape}")
        ...
        lat: shape=(11, 19)
        lon: shape=(11, 19)
        >>> r["lon"][:2]
        array([[-27., -23., -19., -15., -11.,  -7.,  -3.,   1.,   5.,   9.,  13.,
         17.,  21.,  25.,  29.,  33.,  37.,  41.,  45.],
        [-27., -23., -19., -15., -11.,  -7.,  -3.,   1.,   5.,   9.,  13.,
         17.,  21.,  25.,  29.,  33.,  37.,  41.,  45.]])

        """
        if self._is_shared_grid():
            return self[0].to_latlon(**kwargs)
        elif len(self) == 0:
            return dict(lat=None, lon=None)
        else:
            raise ValueError("Fields do not have the same grid geometry")

    def projection(self):
        r"""Return the projection information shared by all the fields.

        Returns
        -------
        :obj:`Projection`

        Raises
        ------
        ValueError
            When not all the fields have the same grid geometry

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> ds.projection()
        <Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to ...>
        Name: unknown
        Axis Info [cartesian]:
        - E[east]: Easting (unknown)
        - N[north]: Northing (unknown)
        - h[up]: Ellipsoidal height (metre)
        Area of Use:
        - undefined
        Coordinate Operation:
        - name: unknown
        - method: Equidistant Cylindrical
        Datum: Unknown based on WGS 84 ellipsoid
        - Ellipsoid: WGS 84
        - Prime Meridian: Greenwich
        >>> ds.projection().to_proj_string()
        '+proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to_meter=111319.4907932736 +no_defs +type=crs'
        """
        if self._is_shared_grid():
            return self[0].projection()
        elif len(self) == 0:
            return None
        else:
            raise ValueError("Fields do not have the same grid geometry")

    def bounding_box(self):
        r"""Return the bounding box for each field.

        Returns
        -------
        list
            List with one :obj:`BoundingBox <data.utils.bbox.BoundingBox>` per
            :obj:`GribField <data.readers.grib.codes.GribField>`
        """
        return [s.bounding_box() for s in self]

    @cached_method
    def _is_shared_grid(self):
        if len(self) > 0:
            grid = self[0].metadata().geography._unique_grid_id()
            if grid is not None:
                return all(
                    f.metadata().geography._unique_grid_id() == grid for f in self
                )
        return False

    def save(self, filename):
        r"""Write all the fields into a file. The target file will be overwritten if
        already exists.

        Parameters
        ----------
        filename: str
            The target file path.
        """
        with open(filename, "wb") as f:
            self.write(f)

    def write(self, f):
        r"""Write all the fields to a file object.

        Parameters
        ----------
        f: file object
            The target file object.
        """
        for s in self:
            s.write(f)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from abc import ABCMeta, abstractmethod


class Geography(metaclass=ABCMeta):
    """Geographical information about a field or data unit"""

    @abstractmethod
    def latitudes(self):
        r"""Return the latitudes.

        Returns
        -------
        ndarray
        """
        pass

    @abstractmethod
    def longitudes(self):
        r"""Return the longitudes.

        Returns
        -------
        ndarray
        """
        pass

    @abstractmethod
    def x(self):
        r"""Return the x coordinates in the original CRS.

        Returns
        -------
        ndarray
        """
        pass

    @abstractmethod
    def y(self):
        r"""Return the y coordinates in the original CRS.

        Returns
        -------
        ndarray
        """
        pass

    @property
    @abstractmethod
    def shape(self):
        r"""Return the shape of the grid or data values.

        Returns
        -------
        tuple
        """
        pass

    @abstractmethod
    def _unique_grid_id(self):
        r"""Return a unique id of the grid of a field."""
        pass

    @abstractmethod
    def projection(self):
        r"""Return information about projection.

        Returns
        -------
        :obj:`Projection`
        """
        pass

    @abstractmethod
    def bounding_box(self):
        r"""Return the bounding box.

        Returns
        -------
        :obj:`BoundingBox <data.utils.bbox.BoundingBox>`
        """
        pass



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
import os
import tempfile

import numpy as np
import pandas as pd

from earthkit.data import from_object, wrappers
from earthkit.data.wrappers import pandas as pd_wrapper

LOG = logging.getLogger(__name__)

TEST_NP = np.arange(10)
TEST_INDEX = np.array([pd.to_datetime(f"2000-01-{i:02d}") for i in range(1, 11)])

TEST_PDS = pd.Series(TEST_NP, name="data", index=TEST_INDEX)
TEST_PDS_LAT = pd.Series(TEST_NP + 50, name="lat", index=TEST_INDEX)
TEST_PDS_LON = pd.Series(TEST_NP - 10, name="lon", index=TEST_INDEX)

TEST_PDDF = pd.concat([TEST_PDS, TEST_PDS_LAT, TEST_PDS_LON], axis=1)


def test_series_wrapper_detection():
    _wrapper = pd_wrapper.wrapper(TEST_PDS)
    assert isinstance(_wrapper, pd_wrapper.PandasSeriesWrapper)
    _wrapper = wrappers.get_wrapper(TEST_PDS)
    assert isinstance(_wrapper, pd_wrapper.PandasSeriesWrapper)
    _wrapper = from_object(TEST_PDS)
    assert isinstance(_wrapper, pd_wrapper.PandasSeriesWrapper)


def test_series_wrapper_methods():
    _wrapper = from_object(TEST_PDS)

    # Check iterable
    for thing in _wrapper:
        assert thing is not None

    _wrapper.describe()

    pd.testing.assert_series_equal(_wrapper.data, _wrapper.to_pandas())

    assert isinstance(_wrapper.to_numpy(), np.ndarray)

    import xarray as xr

    assert isinstance(_wrapper.to_xarray(), xr.DataArray)

    with tempfile.TemporaryDirectory() as tmp_dir:
        fpath = os.path.join(tmp_dir, "tmp.nc")
        _wrapper.to_netcdf(fpath)


def test_dataframe_wrapper_detection():
    _wrapper = pd_wrapper.wrapper(TEST_PDDF)
    assert isinstance(_wrapper, pd_wrapper.PandasDataFrameWrapper)
    _wrapper = wrappers.get_wrapper(TEST_PDDF)
    assert isinstance(_wrapper, pd_wrapper.PandasDataFrameWrapper)
    _wrapper = from_object(TEST_PDDF)
    assert isinstance(_wrapper, pd_wrapper.PandasDataFrameWrapper)


def test_dataframe_wrapper_methods():
    _wrapper = from_object(TEST_PDDF)

    # Check iterable
    for thing in _wrapper:
        assert thing is not None

    _wrapper.describe()

    pd.testing.assert_frame_equal(_wrapper.data, _wrapper.to_pandas())

    assert isinstance(_wrapper.to_numpy(), np.ndarray)

    import xarray as xr

    assert isinstance(_wrapper.to_xarray(), xr.Dataset)

    with tempfile.TemporaryDirectory() as tmp_dir:
        fpath = os.path.join(tmp_dir, "tmp.nc")
        _wrapper.to_netcdf(fpath)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from importlib import import_module

from earthkit.data.core import Base
from earthkit.data.decorators import locked

LOG = logging.getLogger(__name__)


class Wrapper(Base):
    pass


# TODO: Add plugins
def _helpers(function_name, lookup, here=os.path.dirname(__file__), package=__name__):
    if not lookup:
        for path in os.listdir(here):
            if path.endswith(".py") and path[0] not in ("_", "."):
                name, _ = os.path.splitext(path)
                try:
                    lookup[name] = getattr(
                        import_module(f".{name}", package=package), function_name
                    )
                except Exception as e:
                    LOG.warning(f"Error loading {function_name} '{name}': {e}")
    return lookup


_WRAPPERS = {}


@locked
def _wrappers():
    return _helpers("wrapper", _WRAPPERS)


def get_wrapper(data, *args, **kwargs):
    """
    Returns the input object with the appropriate earthkit-data wrapper.
    i.e. so that an xarray object can be handled as an earkit-data object.
    """
    if isinstance(data, Base):
        return data

    for name, h in _wrappers().items():
        wrapper = h(data, *args, **kwargs)
        if wrapper is not None:
            return wrapper.mutate()
    else:
        fullname = ".".join([data.__class__.__module__, data.__class__.__qualname__])
        LOG.warning(
            f"Cannot find a wrapper for class: {fullname}, returning unwrapped object"
        )
        return data


# def from_object(obj: object, *args, **kwargs) -> Base:
#     """
#     Open an object as an earthkit-data object.
#     object type must have a wrapper, otherwise input object is returned.
#     """
#     return get_wrapper(obj, *args, **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from abc import ABCMeta, abstractmethod
from importlib import import_module

from earthkit.data.decorators import locked

LOG = logging.getLogger(__name__)


_WRITERS = {}


class Writer(metaclass=ABCMeta):
    @abstractmethod
    def write(self, f, values, metadata, **kwargs):
        pass


def write(f, values, metadata, **kwargs):
    x = _writers(metadata.__class__.__name__)
    c = x()
    c.write(f, values, metadata, **kwargs)


@locked
def _writers(method_name):
    if not _WRITERS:
        here = os.path.dirname(__file__)
        for path in sorted(os.listdir(here)):
            if path[0] in ("_", "."):
                continue

            if path.endswith(".py") or os.path.isdir(os.path.join(here, path)):
                name, _ = os.path.splitext(path)
                try:
                    module = import_module(f".{name}", package=__name__)
                    if hasattr(module, "Writer"):
                        w = getattr(module, "Writer")
                        _WRITERS[w.METADATA_TYPE] = w
                        _WRITERS[name] = w
                except Exception:
                    LOG.exception("Error loading writer %s", name)

    return _WRITERS[method_name]



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.data import from_source
from earthkit.data.testing import earthkit_file


def test_geojson():
    s = from_source("file", earthkit_file("tests/data/NUTS_RG_20M_2021_3035.geojson"))
    assert s
    for _s in s:
        assert _s is not None
        assert _s.geometry


def test_geojson_bounding_box():
    s = from_source("file", earthkit_file("tests/data/NUTS_RG_20M_2021_3035.geojson"))
    bb = s.bounding_box()
    assert bb
    assert bb.south, bb.north == (-90.0, 90.0)
    assert bb.east, bb.west == (-180.0, 180.0)


# TODO test mimetypes
# def test_csv_mimetypes():
# assert mimetypes.guess_type("x.geojson") == ("application/geo+json", None)
# assert mimetypes.guess_type("x.geojson.gz") == ("application/geo+json", "gzip")
# assert mimetypes.guess_type("x.geojson.bz2") == ("application/geo+json", "bzip2")


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


def test_netcdf_ls():
    f = from_source("file", earthkit_examples_file("tuv_pl.nc"))

    # default keys
    f1 = f[:4]
    df = f1.ls()

    ref = {
        "variable": {0: "t", 1: "t", 2: "t", 3: "t"},
        "level": {0: 1000, 1: 850, 2: 700, 3: 500},
        "time": {
            0: datetime.datetime.fromisoformat("2018-08-01 12:00:00"),
            1: datetime.datetime.fromisoformat("2018-08-01 12:00:00"),
            2: datetime.datetime.fromisoformat("2018-08-01 12:00:00"),
            3: datetime.datetime.fromisoformat("2018-08-01 12:00:00"),
        },
        "units": {0: "K", 1: "K", 2: "K", 3: "K"},
    }

    assert ref == df.to_dict()

    # extra keys
    f1 = f[:2]
    df = f1.ls(extra_keys=["long_name"])

    ref = {
        "variable": {0: "t", 1: "t"},
        "level": {0: 1000, 1: 850},
        "time": {
            0: datetime.datetime.fromisoformat("2018-08-01 12:00:00"),
            1: datetime.datetime.fromisoformat("2018-08-01 12:00:00"),
        },
        "units": {0: "K", 1: "K"},
        "long_name": {0: "Temperature", 1: "Temperature"},
    }
    assert ref == df.to_dict()


if __name__ == "__main__":
    from earthkit.data.testing import main

    # test_datetime()
    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


def test_grib_concat():
    ds1 = from_source("file", earthkit_examples_file("test.grib"))
    ds2 = from_source("file", earthkit_examples_file("test6.grib"))
    ds = ds1 + ds2

    assert len(ds) == 8
    md = ds1.metadata("param") + ds2.metadata("param")
    assert ds.metadata("param") == md


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr

from earthkit.data import from_source, transform, translators, wrappers
from earthkit.data.translators import ndarray as ndtranslator
from earthkit.data.translators import pandas as pdtranslator
from earthkit.data.translators import string as strtranslator
from earthkit.data.translators import xarray as xrtranslator

LOG = logging.getLogger(__name__)


def test_string_translator():
    # Check that an ndarray translator can be created
    _ndwrapper = wrappers.get_wrapper("Eartha")
    _trans = strtranslator.translator(_ndwrapper, str)
    assert isinstance(_trans, strtranslator.StrTranslator)

    # Check that get_translator method find the right translator
    _trans = translators.get_translator("Eartha", str)
    assert isinstance(_trans, strtranslator.StrTranslator)

    # Check that public API method transforms to correct type (back to original)
    transformed = transform("Eartha", str)
    assert isinstance(transformed, str)


def test_ndarray_translator():
    # Check that an ndarray translator can be created
    _ndwrapper = wrappers.get_wrapper(np.array([1]))
    _trans = ndtranslator.translator(_ndwrapper, np.ndarray)
    assert isinstance(_trans, ndtranslator.NumpyNDArrayTranslator)

    # Check that get_translator method find the right translator
    _trans = translators.get_translator(np.array([1]), np.ndarray)
    assert isinstance(_trans, ndtranslator.NumpyNDArrayTranslator)

    # Check that public API method  transforms to correct type (back to original)
    transformed = transform(np.array([1]), np.ndarray)
    assert isinstance(transformed, np.ndarray)


def test_xr_dataarray_translator():
    # Check that an xr.DataArray translator can be created
    _xrwrapper = wrappers.get_wrapper(xr.DataArray())
    _trans = xrtranslator.translator(_xrwrapper, xr.DataArray)
    assert isinstance(_trans, xrtranslator.XArrayDataArrayTranslator)

    # Check that get_translator method find the right translator
    _trans = translators.get_translator(xr.DataArray(), xr.DataArray)
    assert isinstance(_trans, xrtranslator.XArrayDataArrayTranslator)

    # Check that public API method transforms to correct type (back to original)
    transformed = transform(xr.DataArray(), xr.DataArray)
    assert isinstance(transformed, xr.DataArray)


def test_xr_dataset_translator():
    # Check that an xr.Dataset translator can be created
    _xrwrapper = wrappers.get_wrapper(xr.Dataset())
    _trans = xrtranslator.translator(_xrwrapper, xr.Dataset)
    assert isinstance(_trans, xrtranslator.XArrayDatasetTranslator)

    # Check that get_translator method find the right translator
    _trans = translators.get_translator(xr.Dataset(), xr.Dataset)
    assert isinstance(_trans, xrtranslator.XArrayDatasetTranslator)

    # Check that public API method transforms to correct type (back to original)
    transformed = transform(xr.Dataset(), xr.Dataset)
    assert isinstance(transformed, xr.Dataset)


def test_pd_series_translator():
    # Check that an xr.Dataset translator can be created
    _pdwrapper = wrappers.get_wrapper(pd.Series())
    _trans = pdtranslator.translator(_pdwrapper, pd.DataFrame)
    assert isinstance(_trans, pdtranslator.PandasSeriesTranslator)

    # Check that get_translator method find the right translator
    _trans = translators.get_translator(pd.Series(), pd.Series)
    assert isinstance(_trans, pdtranslator.PandasSeriesTranslator)

    # Check that public API method transforms to correct type (back to original)
    transformed = transform(pd.Series(), pd.Series)
    assert isinstance(transformed, pd.Series)


def test_pd_dataframe_translator():
    # Check that an xr.Dataset translator can be created
    _pdwrapper = wrappers.get_wrapper(pd.DataFrame())
    _trans = pdtranslator.translator(_pdwrapper, pd.DataFrame)
    assert isinstance(_trans, pdtranslator.PandasDataFrameTranslator)

    # Check that get_translator method find the right translator
    _trans = translators.get_translator(pd.DataFrame(), pd.DataFrame)
    assert isinstance(_trans, pdtranslator.PandasDataFrameTranslator)

    # Check that public API method transforms to correct type (back to original)
    transformed = transform(pd.DataFrame(), pd.DataFrame)
    assert isinstance(transformed, pd.DataFrame)


def test_gpd_dataframe_translator():
    # Check that an xr.Dataset translator can be created
    _pdwrapper = wrappers.get_wrapper(gpd.GeoDataFrame())
    _trans = pdtranslator.translator(_pdwrapper, gpd.GeoDataFrame)
    assert isinstance(_trans, pdtranslator.GeoPandasDataFrameTranslator)

    # Check that get_translator method find the right translator
    _trans = translators.get_translator(gpd.GeoDataFrame(), gpd.GeoDataFrame)
    assert isinstance(_trans, pdtranslator.GeoPandasDataFrameTranslator)

    # Check that public API method transforms to correct type (back to original)
    transformed = transform(gpd.GeoDataFrame(), gpd.GeoDataFrame)
    assert isinstance(transformed, gpd.GeoDataFrame)


def test_transform_from_grib_file():
    # transform grib-based data object
    f = from_source("file", "tests/data/test_single.grib")

    # np.ndarray
    transformed = transform(f, np.ndarray)
    assert isinstance(transformed, np.ndarray)

    # xr.DataArray
    transformed = transform(f, xr.DataArray)
    assert isinstance(transformed, xr.DataArray)

    # xr.Dataset
    transformed = transform(f, xr.Dataset)
    assert isinstance(transformed, xr.Dataset)


def test_transform_from_xarray_object():
    # transform grib-based data object
    da = xr.DataArray([])
    ds = xr.Dataset({"a": da})

    # da to np.ndarray
    transformed = transform(da, np.ndarray)
    assert isinstance(transformed, np.ndarray)

    # ds to xr.DataArray
    transformed = transform(ds, xr.DataArray)
    assert isinstance(transformed, xr.DataArray)

    # ds to np.ndarray
    transformed = transform(ds, np.ndarray)
    assert isinstance(transformed, np.ndarray)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


def check_array(v, shape=None, first=None, last=None, meanv=None, eps=1e-3):
    assert v.shape == shape
    assert np.isclose(v[0], first, eps)
    assert np.isclose(v[-1], last, eps)
    assert np.isclose(v.mean(), meanv, eps)


def test_netcdf_values_surf():
    f = from_source("file", earthkit_examples_file("test.nc"))

    eps = 1e-5

    # whole file
    v = f.values
    assert isinstance(v, np.ndarray)
    assert v.shape == (2, 209)
    v0 = v[0].flatten()
    check_array(
        v0,
        (209,),
        first=262.78027,
        last=309.62207,
        meanv=283.98642,
        eps=eps,
    )

    v1 = v[1].flatten()
    check_array(
        v1,
        (209,),
        first=101947.81,
        last=100941.875,
        meanv=101201.305,
        eps=eps,
    )

    # field
    v0_f = f[0].values
    assert isinstance(v0_f, np.ndarray)
    assert v0_f.shape == (209,)
    assert np.allclose(v0_f, v0, eps)


def test_netcdf_values_upper():
    f = from_source("file", earthkit_examples_file("tuv_pl.nc"))

    eps = 1e-5

    # whole file
    v = f.values
    assert isinstance(v, np.ndarray)
    assert v.shape == (18, 84)
    vf = v[0].flatten()
    check_array(
        vf,
        (84,),
        first=272.5642,
        last=240.56417846679688,
        meanv=279.70703560965404,
        eps=eps,
    )

    vf = v[15].flatten()
    check_array(
        vf,
        (84,),
        first=-0.9902425,
        last=-8.989977,
        meanv=-2.037789,
        eps=eps,
    )


def test_netcdf_to_numpy_surf():
    f = from_source("file", earthkit_examples_file("test.nc"))

    eps = 1e-5
    v = f.to_numpy()
    assert isinstance(v, np.ndarray)
    assert v.shape == (2, 11, 19)

    v0 = v[0].flatten()
    check_array(
        v0,
        (209,),
        first=262.78027,
        last=309.62207,
        meanv=283.98642,
        eps=eps,
    )

    v1 = v[1].flatten()
    check_array(
        v1,
        (209,),
        first=101947.81,
        last=100941.875,
        meanv=101201.305,
        eps=eps,
    )

    # field
    v0_f = f[0].to_numpy()
    assert isinstance(v0_f, np.ndarray)
    assert v0_f.shape == (11, 19)
    assert np.allclose(v0_f.flatten(), v0, eps)


@pytest.mark.parametrize(
    "first,options, expected_shape",
    [
        (False, {}, (2, 11, 19)),
        (False, {"flatten": True}, (2, 209)),
        (False, {"flatten": False}, (2, 11, 19)),
        (True, {}, (11, 19)),
        (True, {"flatten": True}, (209,)),
        (True, {"flatten": False}, (11, 19)),
    ],
)
def test_netcdf_to_numpy_surf_shape(first, options, expected_shape):
    f = from_source("file", earthkit_examples_file("test.nc"))

    eps = 1e-5

    if first:
        data = f[0]
        v_ref = f[0].values
    else:
        data = f
        v_ref = f.values.flatten()

    v1 = data.to_numpy(**options)
    assert isinstance(v1, np.ndarray)
    assert v1.shape == expected_shape
    v1 = v1.flatten()
    assert np.allclose(v_ref, v1, eps)


@pytest.mark.parametrize(
    "options, expected_shape",
    [
        (
            {},
            (
                18,
                7,
                12,
            ),
        ),
        (
            {"flatten": True},
            (
                18,
                84,
            ),
        ),
        ({"flatten": False}, (18, 7, 12)),
    ],
)
def test_netcdf_to_numpy_upper_shape(options, expected_shape):
    f = from_source("file", earthkit_examples_file("tuv_pl.nc"))

    eps = 1e-5

    # whole file
    v = f.to_numpy()
    assert isinstance(v, np.ndarray)
    assert v.shape == (18, 7, 12)
    vf0 = f[0].to_numpy().flatten()
    assert vf0.shape == (84,)
    vf15 = f[15].to_numpy().flatten()
    assert vf15.shape == (84,)

    v1 = f.to_numpy(**options)
    assert isinstance(v1, np.ndarray)
    assert v1.shape == expected_shape
    vr = v1[0].flatten()
    assert np.allclose(vf0, vr, eps)
    vr = v1[15].flatten()
    assert np.allclose(vf15, vr, eps)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_netcdf_to_numpy_surf_dtype(dtype):
    f = from_source("file", earthkit_examples_file("test.nc"))

    v = f[0].to_numpy(dtype=dtype)
    assert v.dtype == dtype

    v = f.to_numpy(dtype=dtype)
    assert v.dtype == dtype


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_netcdf_to_numpy_upper_dtype(dtype):
    f = from_source("file", earthkit_examples_file("tuv_pl.nc"))

    v = f[0].to_numpy(dtype=dtype)
    assert v.dtype == dtype

    v = f.to_numpy(dtype=dtype)
    assert v.dtype == dtype


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.testing import NO_MARS


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_MARS, reason="No access to MARS")
def test_mars_grib_1():
    s = from_source(
        "mars",
        param=["2t", "msl"],
        levtype="sfc",
        area=[50, -50, 20, 50],
        grid=[2, 2],
        date="2023-05-10",
    )
    assert len(s) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_MARS, reason="No access to MARS")
def test_mars_grib_2():
    s = from_source(
        "mars",
        param=["2t", "msl"],
        levtype="sfc",
        area=[50, -50, 20, 50],
        grid=[2, 2],
        date="2023-05-10",
        split_on="param",
    )
    assert len(s) == 2


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os

import eccodes
import numpy as np

from earthkit.data.core.fieldlist import Field
from earthkit.data.readers.grib.metadata import GribMetadata
from earthkit.data.utils.message import (
    CodesHandle,
    CodesMessagePositionIndex,
    CodesReader,
)

LOG = logging.getLogger(__name__)

_GRIB_NAMESPACES = {"default": None}

for k in ("ls", "geography", "mars", "parameter", "statistics", "time", "vertical"):
    _GRIB_NAMESPACES[k] = k


def missing_is_none(x):
    return None if x == 2147483647 else x


class GribCodesMessagePositionIndex(CodesMessagePositionIndex):
    # This does not belong here, should be in the C library
    def _get_message_positions(self, path):
        fd = os.open(path, os.O_RDONLY)
        try:

            def get(count):
                buf = os.read(fd, count)
                assert len(buf) == count
                return int.from_bytes(
                    buf,
                    byteorder="big",
                    signed=False,
                )

            offset = 0
            while True:
                code = os.read(fd, 4)
                if len(code) < 4:
                    break

                if code != b"GRIB":
                    offset = os.lseek(fd, offset + 1, os.SEEK_SET)
                    continue

                length = get(3)
                edition = get(1)

                if edition == 1:
                    if length & 0x800000:
                        sec1len = get(3)
                        os.lseek(fd, 4, os.SEEK_CUR)
                        flags = get(1)
                        os.lseek(fd, sec1len - 8, os.SEEK_CUR)

                        if flags & (1 << 7):
                            sec2len = get(3)
                            os.lseek(fd, sec2len - 3, os.SEEK_CUR)

                        if flags & (1 << 6):
                            sec3len = get(3)
                            os.lseek(fd, sec3len - 3, os.SEEK_CUR)

                        sec4len = get(3)

                        if sec4len < 120:
                            length &= 0x7FFFFF
                            length *= 120
                            length -= sec4len
                            length += 4

                if edition == 2:
                    length = get(8)

                yield offset, length
                offset = os.lseek(fd, offset + length, os.SEEK_SET)

        finally:
            os.close(fd)


class GribCodesHandle(CodesHandle):
    PRODUCT_ID = eccodes.CODES_PRODUCT_GRIB

    # TODO: just a wrapper around the base class implementation to handle the
    # s,l,d qualifiers. Once these are implemented in the base class this method can
    # be removed. md5GridSection is also handled!
    def get(self, name, ktype=None, **kwargs):
        if name == "values":
            return self.get_values()
        elif name == "md5GridSection":
            return self.get_md5GridSection()

        return super().get(name, ktype, **kwargs)

    def get_md5GridSection(self):
        # Special case because:
        #
        # 1) eccodes is returning size > 1 for 'md5GridSection'
        # (size = 16 : it is the number of bytes of the value)
        # This is already fixed in eccodes 2.27.1
        #
        # 2) sometimes (see below), the value for "shapeOfTheEarth" is inconsistent.
        # This impacts the (computed on-the-fly) value of "md5GridSection".
        # ----------------
        # Example of data with inconsistent values:
        # S2S data, origin='ecmf', param='tp', step='24', number='0', date=['20201203','20200702']
        # the 'md5GridSection' are different
        # This is because one has "shapeOfTheEarth" set to 0, the other to 6.
        # This is only impacting the metadata.
        # Since this has no impact on the data itself,
        # this is unlikely to be fixed. Therefore this hacky patch.
        #
        # Obviously, the patch causes an inconsistency between the value of md5GridSection
        # read by this code, and the value read by another code without this patch.

        save = eccodes.codes_get_long(self._handle, "shapeOfTheEarth")
        eccodes.codes_set_long(self._handle, "shapeOfTheEarth", 255)
        result = eccodes.codes_get_string(self._handle, "md5GridSection")
        eccodes.codes_set_long(self._handle, "shapeOfTheEarth", save)
        return result

    def as_namespace(self, namespace, param="shortName"):
        r = {}
        ignore = {
            "distinctLatitudes",
            "distinctLongitudes",
            "distinctLatitudes",
            "latLonValues",
            "latitudes",
            "longitudes",
            "values",
        }
        for key in self.keys(namespace=namespace):
            if key not in ignore:
                r[key] = self.get(param if key == "param" else key)
        return r

    # TODO: once missing value handling is implemented in the base class this method
    # can be removed
    def get_values(self):
        eccodes.codes_set(self._handle, "missingValue", CodesHandle.MISSING_VALUE)
        vals = eccodes.codes_get_values(self._handle)
        if self.get_long("bitmapPresent"):
            vals[vals == CodesHandle.MISSING_VALUE] = np.nan
        return vals

    def get_latitudes(self):
        return self.get("latitudes")

    def get_longitudes(self):
        return self.get("longitudes")

    def get_data_points(self):
        return eccodes.codes_grib_get_data(self._handle)

    def set_values(self, values):
        try:
            assert self.path is None, "Only cloned handles can have values changed"
            eccodes.codes_set_values(self._handle, values.flatten())
            # This is writing on the GRIB that something has been modified (255=unknown)
            eccodes.codes_set_long(self._handle, "generatingProcessIdentifier", 255)
        except Exception as e:
            LOG.error("Error setting values")
            LOG.exception(e)
            raise


class GribCodesReader(CodesReader):
    PRODUCT_ID = eccodes.CODES_PRODUCT_GRIB
    HANDLE_TYPE = GribCodesHandle


class GribField(Field):
    r"""Represents a GRIB message in a GRIB file.

    Parameters
    ----------
    path: str
        Path to the GRIB file
    offset: number
        File offset of the message (in bytes)
    length: number
        Size of the message (in bytes)
    """

    def __init__(self, path, offset, length):
        super().__init__()
        self.path = path
        self._offset = offset
        self._length = length
        self._handle = None

    @property
    def handle(self):
        r""":class:`CodesHandle`: Gets an object providing access to the low level GRIB message structure."""
        if self._handle is None:
            assert self._offset is not None
            self._handle = GribCodesReader.from_cache(self.path).at_offset(self._offset)
        return self._handle

    @property
    def values(self):
        r"""ndarray: Gets the values stored in the GRIB field as a 1D ndarray."""
        return self.handle.get_values()

    @property
    def offset(self):
        r"""number: Gets the offset (in bytes) of the GRIB field within the GRIB file."""
        if self._offset is None:
            self._offset = int(self.handle.get("offset"))
        return self._offset

    def _make_metadata(self):
        return GribMetadata(self.handle)

    def __repr__(self):
        return "GribField(%s,%s,%s,%s,%s,%s)" % (
            self._metadata.get("shortName", None),
            self._metadata.get("levelist", None),
            self._metadata.get("date", None),
            self._metadata.get("time", None),
            self._metadata.get("step", None),
            self._metadata.get("number", None),
        )

    # def _get(self, name):
    #     """Private, for testing only"""
    #     # paramId is renamed as param to get rid of the
    #     # additional '.128' (in earthkit/data/scripts/grib.py)
    #     if name == "param":
    #         name = "paramId"
    #     return self.handle.get(name)

    def write(self, f):
        r"""Writes the message to a file object.

        Parameters
        ----------
        f: file object
            The target file object.
        """
        # assert isinstance(f, io.IOBase)
        self.handle.write_to(f)

    def message(self):
        r"""Returns a buffer containing the encoded message.

        Returns
        -------
        bytes
        """
        return self.handle.get_buffer()



#!/usr/bin/env python3

# (C) Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file, earthkit_remote_test_data_file


def test_bufr_ls_invalid_num():
    f = from_source("file", earthkit_examples_file("temp_10.bufr"))
    with pytest.raises(ValueError):
        f.ls(n=0)

    with pytest.raises(ValueError):
        f.ls(0)


def test_bufr_ls_num():
    f = from_source("file", earthkit_examples_file("temp_10.bufr"))

    # default keys

    # head
    df = f.ls(n=2)
    ref = {
        "edition": {0: 3, 1: 3},
        "type": {0: 2, 1: 2},
        "subtype": {0: 101, 1: 101},
        "c": {0: 98, 1: 98},
        "mv": {0: 13, 1: 13},
        "lv": {0: 1, 1: 1},
        "subsets": {0: 1, 1: 1},
        "compr": {0: 0, 1: 0},
        "typicalDate": {0: "20081208", 1: "20081208"},
        "typicalTime": {0: "120000", 1: "120000"},
        "ident": {0: "02836", 1: "01400"},
        "lat": {0: 67.37, 1: 56.9},
        "lon": {0: 26.63, 1: 3.35},
    }

    assert ref == df.to_dict()

    # t tail
    df = f.ls(-2)
    ref = {
        "edition": {0: 3, 1: 3},
        "type": {0: 2, 1: 2},
        "subtype": {0: 101, 1: 101},
        "c": {0: 98, 1: 98},
        "mv": {0: 13, 1: 13},
        "lv": {0: 1, 1: 1},
        "subsets": {0: 1, 1: 1},
        "compr": {0: 0, 1: 0},
        "typicalDate": {0: "20081208", 1: "20081208"},
        "typicalTime": {0: "120000", 1: "120000"},
        "ident": {0: "11035", 1: "02963"},
        "lat": {0: 48.25, 1: 60.82},
        "lon": {0: 16.37, 1: 23.5},
    }

    assert ref == df.to_dict()

    df = f.ls(-2)
    assert ref == df.to_dict()


def test_bufr_head_num():
    f = from_source("file", earthkit_examples_file("temp_10.bufr"))

    # default keys
    df = f.head(n=2)
    ref = {
        "edition": {0: 3, 1: 3},
        "type": {0: 2, 1: 2},
        "subtype": {0: 101, 1: 101},
        "c": {0: 98, 1: 98},
        "mv": {0: 13, 1: 13},
        "lv": {0: 1, 1: 1},
        "subsets": {0: 1, 1: 1},
        "compr": {0: 0, 1: 0},
        "typicalDate": {0: "20081208", 1: "20081208"},
        "typicalTime": {0: "120000", 1: "120000"},
        "ident": {0: "02836", 1: "01400"},
        "lat": {0: 67.37, 1: 56.9},
        "lon": {0: 26.63, 1: 3.35},
    }

    assert ref == df.to_dict()

    df = f.head(2)
    assert ref == df.to_dict()


def test_bufr_tail_num():
    f = from_source("file", earthkit_examples_file("temp_10.bufr"))

    # default keys
    df = f.tail(n=2)
    ref = {
        "edition": {0: 3, 1: 3},
        "type": {0: 2, 1: 2},
        "subtype": {0: 101, 1: 101},
        "c": {0: 98, 1: 98},
        "mv": {0: 13, 1: 13},
        "lv": {0: 1, 1: 1},
        "subsets": {0: 1, 1: 1},
        "compr": {0: 0, 1: 0},
        "typicalDate": {0: "20081208", 1: "20081208"},
        "typicalTime": {0: "120000", 1: "120000"},
        "ident": {0: "11035", 1: "02963"},
        "lat": {0: 48.25, 1: 60.82},
        "lon": {0: 16.37, 1: 23.5},
    }

    assert ref == df.to_dict()

    df = f.tail(2)
    assert ref == df.to_dict()


def test_bufr_dump():
    ds = from_source("file", earthkit_examples_file("temp_10.bufr"))
    r = ds[0].dump()
    assert isinstance(r, dict)
    assert "header" in r
    assert "data" in r


@pytest.mark.parametrize(
    "_kwargs,expected_val", [({"subset": 1}, 1), ({}, 1), ({"subset": 2}, 2)]
)
def test_bufr_dump_uncompressed(_kwargs, expected_val):
    ds = from_source(
        "url",
        earthkit_remote_test_data_file(
            "test-data/synop_multi_subset_uncompressed.bufr"
        ),
    )
    n = ds[0].subset_count()
    assert n == 12

    r = ds[0].dump(**_kwargs)
    assert isinstance(r, dict)
    assert "header" in r
    assert "data" in r
    assert r["data"][0] == dict(key="subsetNumber", value=expected_val, units=None)


@pytest.mark.parametrize("_kwargs", [{"subset": 0}, {"subset": None}])
def test_bufr_dump_uncompressed_full(_kwargs):
    ds = from_source(
        "url",
        earthkit_remote_test_data_file(
            "test-data/synop_multi_subset_uncompressed.bufr"
        ),
    )

    n = ds[0].subset_count()
    assert n == 12

    r = ds[0].dump(**_kwargs)
    assert isinstance(r, dict)
    assert "header" in r
    assert "data" in r
    d = r["data"]
    for i in range(n):
        assert d[i][0] == dict(key="subsetNumber", value=i + 1, units=None)


@pytest.mark.parametrize(
    "_kwargs, expected_val",
    [
        ({"subset": 0}, "[0, ...] (51 items)"),
        ({"subset": None}, "[0, ...] (51 items)"),
        ({"subset": 1}, "0 (51 items)"),
        ({}, "0 (51 items)"),
        ({"subset": 2}, "1 (51 items)"),
    ],
)
def test_bufr_dump_compressed(_kwargs, expected_val):
    ds = from_source(
        "url",
        earthkit_remote_test_data_file("test-data/ens_multi_subset_compressed.bufr"),
    )
    n = ds[0].subset_count()
    assert n == 51

    r = ds[0].dump(**_kwargs)
    assert isinstance(r, dict)
    assert "header" in r
    assert "data" in r
    assert r["data"][1] == dict(
        key="ensembleMemberNumber", value=expected_val, units="Numeric"
    )


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import cdsapi
import yaml

from .cds import CdsRetriever
from .prompt import APIKeyPrompt


class ADSAPIKeyPrompt(APIKeyPrompt):
    register_or_sign_in_url = "https://ads.atmosphere.copernicus.eu/"
    retrieve_api_key_url = "https://ads.atmosphere.copernicus.eu/api-how-to"

    prompts = [
        dict(
            name="url",
            default="https://ads.atmosphere.copernicus.eu/api/v2",
            title="API url",
            validate=r"http.?://.*",
        ),
        dict(
            name="key",
            example="123:abcdef01-0000-1111-2222-0123456789ab",
            title="API key",
            hidden=True,
            validate=r"\d+:[\-0-9a-f]+",
        ),
    ]

    rcfile = "~/.adsapirc"

    def save(self, input, file):
        yaml.dump(input, file, default_flow_style=False)


def client():
    prompt = ADSAPIKeyPrompt()
    prompt.check()

    path = os.path.expanduser(prompt.rcfile)

    if not os.path.exists(path):
        prompt.ask_user_and_save()

    with open(path) as f:
        rc = yaml.safe_load(f.read())

    return cdsapi.Client(**rc)


EXTENSIONS = {
    "grib": ".grib",
    "netcdf": ".nc",
}


class AdsRetriever(CdsRetriever):
    sphinxdoc = """
    AdsRetriever
    """

    def client(self):
        return client()


source = AdsRetriever



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.utils.serialise import (
    SERIALISATION,
    deserialise_state,
    serialise_state,
)

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from indexing_fixtures import get_tmp_fixture  # noqa E402


# TODO: add test for mode=multi
@pytest.mark.parametrize("mode", ["file", "directory"])
@pytest.mark.parametrize("params", (["t", "u"], ["u", "t"]))
@pytest.mark.parametrize("levels", ([500, 850], [850, 500]))
def test_indexing_pickle(mode, params, levels):
    tmp, path = get_tmp_fixture(mode)
    ds = from_source("file", path, indexing=True)

    request = dict(
        level=levels,
        variable=params,
        date=20180801,
        time="1200",
    )

    ds = ds.sel(**request)
    ds = ds.order_by(["level", "param"])

    assert len(ds) == 4, (len(ds), ds, SERIALISATION)
    state = serialise_state(ds)
    ds = deserialise_state(state)
    assert len(ds) == 4, (len(ds), ds, SERIALISATION)

    ref = dict(shortName=["t", "u", "t", "u"], level=[500, 500, 850, 850])

    for k, v in ref.items():
        assert ds.metadata(k) == v


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

from earthkit.data import download_example_file, remote_example_file


def test_download_example_file_single(tmpdir):
    filename = "test.grib"
    with tmpdir.as_cwd():
        download_example_file(filename)
        path = os.path.join(tmpdir.realpath(), filename)
        assert os.path.exists(path)

        mtime = os.stat(path).st_mtime_ns
        download_example_file(filename, force=True)
        assert os.path.exists(path)
        assert os.stat(path).st_mtime_ns > mtime


def test_download_example_file_multi(tmpdir):
    filenames = ["test.grib", "test6.grib"]
    with tmpdir.as_cwd():
        download_example_file(filenames)

        paths = [os.path.join(tmpdir.realpath(), f) for f in filenames]
        for path in paths:
            assert os.path.exists(path)

        mtimes = [os.stat(p).st_mtime_ns for p in paths]
        download_example_file(filenames, force=True)
        for i, p in enumerate(paths):
            assert os.path.exists(p)
            assert os.stat(p).st_mtime_ns > mtimes[i]


def test_remote_example_file_path():
    filename = "test.grib"
    assert (
        remote_example_file(filename)
        == "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/"
        + filename
    )


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file, earthkit_test_data_file
from earthkit.data.utils import projections


def check_array(v, shape=None, first=None, last=None, meanv=None, eps=1e-3):
    assert v.shape == shape
    assert np.isclose(v[0], first, eps)
    assert np.isclose(v[-1], last, eps)
    assert np.isclose(v.mean(), meanv, eps)


@pytest.mark.parametrize("index", [0, None])
def test_grib_to_latlon_single(index):
    f = from_source("file", earthkit_test_data_file("test_single.grib"))

    eps = 1e-5
    g = f[index] if index is not None else f
    v = g.to_latlon(flatten=True)
    assert isinstance(v, dict)
    assert isinstance(v["lon"], np.ndarray)
    assert isinstance(v["lat"], np.ndarray)
    check_array(
        v["lon"],
        (84,),
        first=0.0,
        last=330.0,
        meanv=165.0,
        eps=eps,
    )
    check_array(
        v["lat"],
        (84,),
        first=90,
        last=-90,
        meanv=0,
        eps=eps,
    )


@pytest.mark.parametrize("index", [0, None])
def test_grib_to_latlon_single_shape(index):
    f = from_source("file", earthkit_test_data_file("test_single.grib"))

    g = f[index] if index is not None else f
    v = g.to_latlon()
    assert isinstance(v, dict)
    assert isinstance(v["lon"], np.ndarray)
    assert isinstance(v["lat"], np.ndarray)

    # x
    assert v["lon"].shape == (7, 12)
    for x in v["lon"]:
        assert np.allclose(x, np.linspace(0, 330, 12))

    # y
    assert v["lat"].shape == (7, 12)
    for i, y in enumerate(v["lat"]):
        assert np.allclose(y, np.ones(12) * (90 - i * 30))


def test_grib_to_latlon_multi():
    f = from_source("file", earthkit_examples_file("test.grib"))

    v_ref = f[0].to_latlon(flatten=True)
    v = f.to_latlon(flatten=True)
    assert isinstance(v, dict)
    assert v.keys() == v_ref.keys()

    assert isinstance(v, dict)
    assert np.allclose(v["lat"], v_ref["lat"])
    assert np.allclose(v["lon"], v_ref["lon"])


def test_grib_to_latlon_multi_non_shared_grid():
    f1 = from_source("file", earthkit_examples_file("test.grib"))
    f2 = from_source("file", earthkit_examples_file("test4.grib"))
    f = f1 + f2

    with pytest.raises(ValueError):
        f.to_latlon()


@pytest.mark.parametrize("index", [0, None])
def test_grib_to_points_single(index):
    f = from_source("file", earthkit_test_data_file("test_single.grib"))

    eps = 1e-5
    g = f[index] if index is not None else f
    v = g.to_points(flatten=True)
    assert isinstance(v, dict)
    assert isinstance(v["x"], np.ndarray)
    assert isinstance(v["y"], np.ndarray)
    check_array(
        v["x"],
        (84,),
        first=0.0,
        last=330.0,
        meanv=165.0,
        eps=eps,
    )
    check_array(
        v["y"],
        (84,),
        first=90,
        last=-90,
        meanv=0,
        eps=eps,
    )


def test_grib_to_points_unsupported_grid():
    f = from_source("file", earthkit_test_data_file("mercator.grib"))
    with pytest.raises(ValueError):
        f[0].to_points()


def test_grib_to_points_multi():
    f = from_source("file", earthkit_examples_file("test.grib"))

    v_ref = f[0].to_points(flatten=True)
    v = f.to_points(flatten=True)
    assert isinstance(v, dict)
    assert v.keys() == v_ref.keys()

    assert isinstance(v, dict)
    assert np.allclose(v["x"], v_ref["x"])
    assert np.allclose(v["y"], v_ref["y"])


def test_grib_to_points_multi_non_shared_grid():
    f1 = from_source("file", earthkit_examples_file("test.grib"))
    f2 = from_source("file", earthkit_examples_file("test4.grib"))
    f = f1 + f2

    with pytest.raises(ValueError):
        f.to_points()


def test_bbox():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    bb = ds.bounding_box()
    assert len(bb) == 2
    for b in bb:
        assert b.as_tuple() == (73, -27, 33, 45)


@pytest.mark.parametrize("index", [0, None])
def test_grib_projection_ll(index):
    f = from_source("file", earthkit_examples_file("test.grib"))

    if index is not None:
        g = f[index]
    else:
        g = f
    assert isinstance(
        g.projection(), (projections.EquidistantCylindrical, projections.LongLat)
    )


def test_grib_projection_mercator():
    f = from_source("file", earthkit_test_data_file("mercator.grib"))
    projection = f[0].projection()
    assert isinstance(projection, projections.Mercator)
    assert projection.parameters == {
        "true_scale_latitude": 20,
        "central_latitude": 0,
        "central_longitude": 0,
        "false_easting": 0,
        "false_northing": 0,
    }
    assert projection.globe == dict()


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
from earthkit.data import from_source
from earthkit.data.decorators import normalize
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.utils.bbox import BoundingBox


@normalize("area", "bounding-box")
def bbox_list(ignore, area):
    return area


@normalize("area", "bounding-box(tuple)")
def bbox_tuple(area, ignore=None):
    return area


@normalize("area", "bounding-box(list)")
def bbox_bbox(area):
    return area


@normalize("area", "bounding-box(dict)")
def bbox_dict(area):
    return area


@normalize("area", "bounding-box")
def bbox_defaults(area=None):
    return area


def test_bbox():
    area = [30.0, 2.0, 3.0, 4.0]
    bbox = BoundingBox(north=30, west=2, south=3, east=4)

    assert bbox_list(None, area) == bbox
    assert bbox_list(area=area, ignore=None) == bbox

    assert bbox_tuple(area) == tuple(area)
    assert bbox_tuple(area=area) == tuple(area)

    assert bbox_bbox(area) == area

    assert bbox_dict(area) == dict(north=30, west=2, south=3, east=4)

    assert bbox_defaults(area) == bbox

    source = from_source("file", earthkit_examples_file("test.grib"))
    assert bbox_tuple(source[0]) == (73.0, -27.0, 33.0, 45.0)

    source = from_source("file", earthkit_examples_file("test.nc"))
    assert bbox_tuple(source[0]) == (73.0, -27.0, 33.0, 45.0)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

from earthkit.data.utils.bbox import BoundingBox
from earthkit.data.wrappers import Wrapper

logger = logging.getLogger(__name__)


class PandasSeriesWrapper(Wrapper):
    """Wrapper around a `pandas.DataFrame`, offering polymorphism and
    convenience methods.
    """

    def __init__(self, data):
        self.data = data

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, i):
        return self.data.columns[i]

    def __len__(self):
        return len(self.data)

    def to_pandas(self):
        """
        Return a `pandas.Series` representation of the data.

        Returns
        -------
        pandas.Series
        """
        return self.data

    def to_numpy(self, **kwargs):
        """
        Return a `numpy.array` representation of the data.

        Returns
        -------
        numpy.array
        """
        return self.data.to_numpy(**kwargs)

    def to_xarray(self, **kwargs):
        """
        Return a `xarray.DataArray` representation of the data.

        Returns
        -------
        numpy.array
        """
        return self.data.to_xarray(**kwargs)

    def to_netcdf(self, *args, **kwargs):
        """
        Write the pandas object to a netCDF file

        Returns
        -------
        numpy.array
        """
        return self.data.to_xarray().to_netcdf(*args, **kwargs)

    def to_json(self, *args, **kwargs):
        """
        Write the pandas object to a netCDF file

        Returns
        -------
        numpy.array
        """
        kwargs.setdefault("indent", 2)
        return self.data.to_json(*args, **kwargs)

    def describe(self):
        """
        A pandas is tabular, therefore return itself

        Returns
        -------
        numpy.array
        """
        return self.data


class PandasDataFrameWrapper(PandasSeriesWrapper):
    """Wrapper around a `pandas.DataFrame`, offering polymorphism and
    convenience methods.
    """

    def to_pandas(self):
        """
        Return a `pandas.DataFrame` representation of the data.

        Returns
        -------
        pandas.DataFrame
        """
        return self.data

    def to_xarray(self, **kwargs):
        """
        Return a `xarray.Dataset` representation of the data.

        Returns
        -------
        numpy.array
        """
        return self.data.to_xarray(**kwargs)


class GeoPandasDataFrameWrapper(PandasDataFrameWrapper):
    """Wrapper around a `geopandas.DataFrame`, offering polymorphism and
    convenience methods.

    Key difference to a pandas dataframe wrapper, we treat rows (features) as fields
    """

    def __init__(self, source):
        super().__init__(source)
        try:
            (
                self.east,
                self.south,
                self.west,
                self.north,
            ) = self.data.crs.area_of_use.bounds
        except AttributeError:
            logger.warn("Bounding box not found in geopandas object")
        self.fields = None

    def __iter__(self):
        """
        Iterate over features in geojson via pandas
        """
        self._scan()
        return iter(self.fields)

    def __getitem__(self, i):
        self._scan()
        return self.fields[i]

    def __len__(self):
        return len(self.data)

    def _scan(self):
        if self.fields is None:
            self.fields = self.get_fields()

    def get_fields(self):
        """
        For geopandas, a field is a feature
        """
        return [row[1] for row in self.data.iterrows()]

    def bounding_box(self):
        """
        For geopandas, get bounding box and convert to EK.BoundingBox type
        """
        return BoundingBox(
            north=self.north, south=self.south, east=self.east, west=self.west
        )


def wrapper(data, *args, **kwargs):
    import pandas as pd

    try:
        import geopandas as gpd

        l_gpd = True
    except ImportError:
        l_gpd = False

    if l_gpd and isinstance(data, gpd.geodataframe.GeoDataFrame):
        return GeoPandasDataFrameWrapper(data, *args, **kwargs)
    if isinstance(data, pd.DataFrame):
        return PandasDataFrameWrapper(data, *args, **kwargs)
    if isinstance(data, pd.Series):
        return PandasSeriesWrapper(data, *args, **kwargs)

    return None



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import operator


class Interval:
    def __init__(self, left, right, closed="both"):
        if left is not None and right is not None and left >= right:
            raise ValueError(f"Invalid bounds, left={left} >= right={right}")
        if closed not in ("both", "left", "right", "none"):
            raise ValueError(f"Invalid closed={closed}")

        self._left = left
        self._right = right
        self._closed = closed
        self._left_oper = operator.lt if closed in ["both", "left"] else operator.le
        self._right_oper = operator.gt if closed in ["both", "right"] else operator.ge

    def __contains__(self, value):
        if self.bounded:
            if self._left is not None and self._left_oper(value, self._left):
                return False
            elif self._right is not None and self._right_oper(value, self._right):
                return False
        return True

    def __repr__(self):
        return f"{self.__class__.__name__}(left={self.left}, right={self.right}, closed={self.closed})"

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    @property
    def closed(self):
        return self._closed

    @property
    def closed_left(self):
        return self._closed in ["both", "left"]

    @property
    def closed_right(self):
        return self._closed in ["both", "right"]

    @property
    def open_left(self):
        return not self.closed_left

    @property
    def open_right(self):
        return not self.closed_right

    @property
    def bounded_left(self):
        return self.left is not None

    @property
    def bounded_right(self):
        return self.right is not None

    @property
    def bounded(self):
        return self.bounded_left or self.bounded_right



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import re
import sys

import pytest

from earthkit.data.testing import MISSING, earthkit_file

# See https://www.blog.pythonlibrary.org/2018/10/16/testing-jupyter-notebooks/


EXAMPLES = earthkit_file("docs", "examples")

SKIP = ["fdb.ipynb", "mars.ipynb", "cds.ipynb", "ads.ipynb"]


def notebooks_list():
    notebooks = []
    for path in os.listdir(EXAMPLES):
        if re.match(r".+\.ipynb$", path):
            # if re.match(r"^\d\d-.*\.ipynb$", path):
            if "Copy" not in path:
                notebooks.append(path)

    return sorted(notebooks)


@pytest.mark.notebook
@pytest.mark.skipif(
    MISSING("nbformat", "nbconvert", "ipykernel"),
    reason="python package nbformat not installed",
)
# @pytest.mark.skipif(not IN_GITHUB, reason="Not on GITHUB")
@pytest.mark.skipif(
    sys.platform == "win32", reason="Cannot execute notebooks on Windows"
)
@pytest.mark.parametrize("path", notebooks_list())
def test_notebook(path):
    import nbformat
    from nbconvert.preprocessors import ExecutePreprocessor

    if path in SKIP:
        pytest.skip("Notebook marked as 'skip'")

    with open(os.path.join(EXAMPLES, path)) as f:
        nb = nbformat.read(f, as_version=4)

    proc = ExecutePreprocessor(timeout=60 * 60 * 5, kernel_name="python3")
    proc.preprocess(nb, {"metadata": {"path": EXAMPLES}})


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# The code is copied from skinnywms, and we should combile later

import mimetypes

import numpy as np

from earthkit.data.readers import Reader
from earthkit.data.wrappers.pandas import GeoPandasDataFrameWrapper


class GeojsonReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)
        self.fields = None

    def _scan(self):
        """
        For geojson, a field is a feature
        """
        if self.fields is None:
            self.fields = self.get_fields()

    def __repr__(self):
        return f"GeojsonReader({self.path})"
        # return self.to_pandas().__repr__()

    def _repr_html_(self):
        html_repr = (
            f"<h3>GeojsonReader(represented as a geopandas object):</h3>"
            f"{self.to_pandas()._repr_html_()}"
        )

        return html_repr

    def __iter__(self):
        """
        Iterate over features in geojson via pandas
        """
        self._scan()
        return iter(self.fields)

    def __len__(self):
        self._scan()
        return len(self.fields)

    def __getitem__(self, n):
        self._scan()
        return self.fields[n]

    def mutate_source(self):
        # A Geojson is a source itself
        return self

    def bounding_box(self, **kwargs):
        return self._to_geopandas_dataframe_wrapper(**kwargs).bounding_box()

    def ls(self, **kwargs):
        return self.to_pandas(**kwargs)

    def describe(self, **kwargs):
        return self.to_pandas(**kwargs)

    def get_fields(self, **kwargs):
        return [row[1] for row in self.to_pandas(**kwargs).iterrows()]

    def to_numpy(self, flatten=False):
        arr = self.to_pandas().to_numpy()
        if flatten:
            arr = arr.flatten()
        return arr

    def to_pandas(self, **kwargs):
        # TODO: handle multiple paths
        return self.to_pandas_from_multi_paths([self.path], **kwargs)

    def to_geopandas(self, **kwargs):
        # TODO: handle multiple paths
        return self.to_pandas(**kwargs)

    def to_xarray(self, **kwargs):
        return self.to_pandas(**kwargs).to_xarray()

    @classmethod
    def to_pandas_from_multi_paths(cls, paths, **kwargs):
        import geopandas as gpd

        geo_df = gpd.pd.concat([gpd.read_file(path, **kwargs) for path in paths])

        return geo_df.set_index(np.arange(len(geo_df)))

    def _to_geopandas_dataframe_wrapper(self, **kwargs):
        return GeoPandasDataFrameWrapper(self.to_pandas(**kwargs))


def reader(source, path, magic=None, deeper_check=False):
    kind, compression = mimetypes.guess_type(path)
    ext = path.split(".")[-1]

    geojson_extensions = ["geojson"]
    geojson_mimetypes = ["application/geo+json"]
    if magic is None or ext in geojson_extensions or kind in geojson_mimetypes:
        return GeojsonReader(source, path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import functools
import logging
import math
from abc import abstractmethod
from collections import defaultdict

import numpy as np

import earthkit.data
from earthkit.data.core.order import build_remapping, normalize_order_by
from earthkit.data.core.select import normalize_selection, selection_from_index
from earthkit.data.sources import Source

LOG = logging.getLogger(__name__)


class OrderOrSelection:
    actions = {}

    def __str__(self):
        return f"{self.__class__.__name__}({self.actions})"

    @property
    def is_empty(self):
        return not self.actions


class Selection(OrderOrSelection):
    def __init__(self, kwargs, remapping=None):
        self.remapping = build_remapping(remapping)

        class InList:
            def __init__(self, lst):
                self.first = True
                self.lst = lst  # lazy casting: lst will be modified

            def __call__(self, x):
                if self.first and x is not None:
                    self.lst = [
                        type(x) if not type(x) is type(y) else y for y in self.lst
                    ]
                    self.first = False
                return x in self.lst

        class InSlice:
            def __init__(self, slc):
                self.slc = slc
                if self.slc.start is None and self.slc.stop is None:
                    raise ValueError("Invalid selection value: slice(None, None)")

                if (
                    self.slc.start is not None
                    and self.slc.stop is not None
                    and self.slc.stop < self.slc.start
                ):
                    self.slc = slice(self.slc.stop, self.slc.start)

            def __call__(self, x):
                return not (
                    (self.slc.start is not None and x < self.slc.start)
                    or (self.slc.stop is not None and x > self.slc.stop)
                )

        self.actions = {}
        for k, v in kwargs.items():
            if v is None or v is earthkit.data.ALL:
                self.actions[k] = lambda x: True
                continue

            if callable(v):
                self.actions[k] = v
                continue

            if isinstance(v, slice):
                self.actions[k] = InSlice(v)
                continue

            if not isinstance(v, (list, tuple, set)):
                v = [v]

            v = set(v)

            self.actions[k] = InList(v)

    def match_element(self, element):
        metadata = self.remapping(element.metadata)
        return all(v(metadata(k, default=None)) for k, v in self.actions.items())


class OrderBase(OrderOrSelection):
    def __init__(self, kwargs, remapping):
        self.actions = self.build_actions(kwargs)
        self.remapping = remapping

    @abstractmethod
    def build_actions(self, kwargs):
        raise NotImplementedError()

    def compare_elements(self, a, b):
        a_metadata = self.remapping(a.metadata)
        b_metadata = self.remapping(b.metadata)
        for k, v in self.actions.items():
            n = v(a_metadata(k, default=None), b_metadata(k, default=None))
            if n != 0:
                return n
        return 0


class Order(OrderBase):
    def build_actions(self, kwargs):
        actions = {}

        def ascending(a, b):
            if a == b:
                return 0
            if a > b:
                return 1
            if a < b:
                return -1
            raise ValueError(f"{a},{b}")

        def descending(a, b):
            if a == b:
                return 0
            if a > b:
                return -1
            if a < b:
                return 1
            raise ValueError(f"{a},{b}")

        class Compare:
            def __init__(self, order):
                self.order = order

            def __call__(self, a, b):
                return ascending(self.get(a), self.get(b))

            def get(self, x):
                return self.order[x]

        for k, v in kwargs.items():
            if v == "ascending" or v is None:
                actions[k] = ascending
                continue

            if v == "descending":
                actions[k] = descending
                continue

            if callable(v):
                actions[k] = v
                continue

            assert isinstance(
                v, (list, tuple)
            ), f"Invalid argument for {k}: {v} ({type(v)})"

            order = {}
            for i, key in enumerate(v):
                order[str(key)] = i
                try:
                    order[int(key)] = i
                except ValueError:
                    pass
                try:
                    order[float(key)] = i
                except ValueError:
                    pass
            actions[k] = Compare(order)

        return actions


class Index(Source):
    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return MaskIndex(*args, **kwargs)

    @abstractmethod
    def __len__(self):
        self._not_implemented()

    def _normalize_kwargs_names(self, **kwargs):
        return kwargs

    def sel(self, *args, remapping=None, **kwargs):
        """Uses metadata values to select a subset of the elements from a fieldlist-like object.

        Parameters
        ----------
        *args: tuple
            Positional arguments specifying the filter condition as dict.
            (See below for details).
        remapping: dict
            Creates new metadata keys from existing ones that we can refer to in ``*args`` and
            ``**kwargs``. E.g. to define a new
            key "param_level" as the concatenated value of the "param" and "level" keys use::

                remapping={"param_level": "{param}{level}"}

            See below for a more elaborate example.

        **kwargs: dict, optional
            Other keyword arguments specifying the filter conditions.
            (See below for details).

        Returns
        -------
        object
            Returns a new object with the filtered elements. It contains a view to the data in the
            original object, so no data is copied.


        Filter conditions are specified by a set of **metadata** keys either by a dictionary (in
        ``*args``) or a set of ``**kwargs``. Both single or multiple keys are allowed to use and each
        can specify the following type of filter values:

        - single value::

            ds.sel(param="t")

        - list of values::

            ds.sel(param=["u", "v"])

        - **slice** of values (defines a **closed interval**, so treated as inclusive of both the start
        and stop values, unlike normal Python indexing)::

            # filter levels between 300 and 500 inclusively
            ds.sel(level=slice(300, 500))

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/tuv_pl.grib")
        >>> len(ds)
        18

        Selecting by a single key ("param"):

        >>> subset = ds.sel(param="t")
        >>> for f in subset:
        ...     print(f)
        ...
        GribField(t,1000,20180801,1200,0,0)
        GribField(t,850,20180801,1200,0,0)
        GribField(t,700,20180801,1200,0,0)
        GribField(t,500,20180801,1200,0,0)
        GribField(t,400,20180801,1200,0,0)
        GribField(t,300,20180801,1200,0,0)


        Selecting by multiple keys ("param", "level") with a list and slice of values:

        >>> subset = ds.sel(param=["u", "v"], level=slice(400, 700))
        >>> for f in subset:
        ...     print(f)
        ...
        GribField(u,700,20180801,1200,0,0)
        GribField(v,700,20180801,1200,0,0)
        GribField(u,500,20180801,1200,0,0)
        GribField(v,500,20180801,1200,0,0)
        GribField(u,400,20180801,1200,0,0)
        GribField(v,400,20180801, 1200,0,0)

        Using ``remapping`` to specify the selection by a key created from two other keys
        (we created key "param_level" from "param" and "levelist"):

        >>> for f in ds.order_by(
        ...     param_level=["t850", "u1000"],
        ...     remapping={"param_level": "{param}{levelist}"},
        ... ):
        ...     print(f)
        GribField(t,850,20180801,1200,0,0)
        GribField(u,1000,20180801,1200,0,0)
        """
        kwargs = normalize_selection(*args, **kwargs)
        kwargs = self._normalize_kwargs_names(**kwargs)
        if not kwargs:
            return self

        selection = Selection(kwargs, remapping=remapping)
        if selection.is_empty:
            return self

        indices = (
            i for i, element in enumerate(self) if selection.match_element(element)
        )

        return self.new_mask_index(self, indices)

    def isel(self, *args, **kwargs):
        """Uses metadata value indices to select a subset of the elements from a
        fieldlist-like object.

        Parameters
        ----------
        *args: tuple
            Positional arguments specifying the filter conditions.
            (See below for details).
        **kwargs: dict, optional
            Other keyword arguments specifying the metadata keys to perform the filtering on.
            (See below for details).

        Returns
        -------
        object
            Returns a new object with the filtered elements. It contains a view to the data in
            the original object, so no data is copied.


        :obj:`isel` works similarly to :obj:`sel` but conditions are specified by indices of metadata
        keys. A metadata index stores the unique, **sorted** values of the corresponding metadata key
        from all the fields in the input data. If the object is a
        obj:`FieldList <data.readers.grib.index.FieldList>`
        to list the indices that have more than one values use
        :meth:`FieldList.indices() <data.readers.grib.index.FieldList.indices>`, or to find
        out the values of a specific index use :meth:`FieldList.index()
        <data.readers.grib.index.FieldList.index>`.

        Filter conditions are specified by a set of **metadata** keys either by a dictionary (in
        ``*args``) or a set of ``**kwargs``. Both single or multiple keys are allowed to use and each
        can specify the following type of filter values:

        - single index::

            ds.sel(param=1)

        - list of indices::

            ds.sel(param=[1, 3])

        - **slice** of values (behaves like normal Python indexing, stop value not included)::

            # filter levels on level indices 1 and 2
            ds.sel(level=slice(1,3))

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/tuv_pl.grib")

        >>> len(ds)
        18
        >>> ds.indices
        {'levelist': (1000, 850, 700, 500, 400, 300), 'param': ('t', 'u', 'v')}

        >>> subset = ds.isel(param=0)
        >>> len(ds)
        6

        >>> for f in subset:
        ...     print(f)
        ...
        GribField(t,1000,20180801,1200,0,0)
        GribField(t,850,20180801,1200,0,0)
        GribField(t,700,20180801,1200,0,0)
        GribField(t,500,20180801,1200,0,0)
        GribField(t,400,20180801,1200,0,0)
        GribField(t,300,20180801,1200,0,0)

        >>> subset = ds.isel(param=[1, 2], level=slice(2, 4))
        >>> len(subset)
        4

        >>> for f in subset:
        ...     print(f)
        ...
        GribField(u,700,20180801,1200,0,0)
        GribField(v,700,20180801,1200,0,0)
        GribField(u,500,20180801,1200,0,0)
        GribField(v,500,20180801,1200,0,0)

        """
        kwargs = normalize_selection(*args, **kwargs)
        kwargs = self._normalize_kwargs_names(**kwargs)
        if not kwargs:
            return self

        kwargs = selection_from_index(self.index, kwargs)

        if not kwargs:
            return self.new_mask_index(self, [])

        return self.sel(**kwargs)

    def order_by(self, *args, remapping=None, **kwargs):
        """Changes the order of the elements in a fieldlist-like object.

        Parameters
        ----------
        *args: tuple
            Positional arguments specifying the metadata keys to perform the ordering on.
            (See below for details)
        remapping: dict
            Defines new metadata keys from existing ones that we can refer to in ``*args`` and
            ``**kwargs``. E.g. to define a new
            key "param_level" as the concatenated value of the "param" and "level" keys use::

                remapping={"param_level": "{param}{level}"}

            See below for a more elaborate example.

        **kwargs: dict, optional
            Other keyword arguments specifying the metadata keys to perform the ordering on.
            (See below for details)

        Returns
        -------
        object
            Returns a new object with reordered elements. It contains a view to the data in the
            original object, so no data is copied.


        Ordering by a single metadata key ("param"). The default ordering direction
        is ``ascending``:

        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test6.grib")
        >>> for f in ds.order_by("param"):
        ...     print(f)
        ...
        GribField(t,850,20180801,1200,0,0)
        GribField(t,1000,20180801,1200,0,0)
        GribField(u,850,20180801,1200,0,0)s
        GribField(u,1000,20180801,1200,0,0)
        GribField(v,850,20180801,1200,0,0)
        GribField(v,1000,20180801,1200,0,0)

        Ordering by multiple keys (first by "level" then by "param"):

        >>> for f in ds.order_by(["level", "param"]):
        ...     print(f)
        ...
        GribField(t,850,20180801,1200,0,0)
        GribField(u,850,20180801,1200,0,0)
        GribField(v,850,20180801,1200,0,0)
        GribField(t,1000,20180801,1200,0,0)
        GribField(u,1000,20180801,1200,0,0)
        GribField(v,1000,20180801,1200,0,0)

        Specifying the ordering direction:

        >>> for f in ds.order_by(param="ascending", level="descending"):
        ...     print(f)
        ...
        GribField(t,1000,20180801,1200,0,0)
        GribField(t,850,20180801,1200,0,0)
        GribField(u,1000,20180801,1200,0,0)
        GribField(u,850,20180801,1200,0,0)
        GribField(v,1000,20180801,1200,0,0)
        GribField(v,850,20180801,1200,0,0)

        Using the list of all the values of a key ("param") to define the order:

        >>> for f in ds.order_by(param=["u", "t", "v"]):
        ...     print(f)
        ...
        GribField(u,1000,20180801,1200,0,0)
        GribField(u,850,20180801,1200,0,0)
        GribField(t,1000,20180801,1200,0,0)
        GribField(t,850,20180801,1200,0,0)
        GribField(v,1000,20180801,1200,0,0)
        GribField(v,850,20180801,1200,0,0)

        Using ``remapping`` to specify the order by a key created from two other keys
        (we created key "param_level" from "param" and "levelist"):

        >>> ordering = ["t850", "t1000", "u1000", "v850", "v1000", "u850"]
        >>> for f in ds.order_by(
        ...     param_level=ordering, remapping={"param_level": "{param}{levelist}"}
        ... ):
        ...     print(f)
        GribField(t,850,20180801,1200,0,0)
        GribField(t,1000,20180801,1200,0,0)
        GribField(u,1000,20180801,1200,0,0)
        GribField(v,850,20180801,1200,0,0)
        GribField(v,1000,20180801,1200,0,0)
        GribField(u,850,20180801,1200,0,0)
        """
        kwargs = normalize_order_by(*args, **kwargs)
        kwargs = self._normalize_kwargs_names(**kwargs)

        remapping = build_remapping(remapping)

        if not kwargs:
            return self

        order = Order(kwargs, remapping=remapping)
        # order = Order(*args, **kwargs)
        if order.is_empty:
            return self

        def cmp(i, j):
            return order.compare_elements(self[i], self[j])

        indices = list(range(len(self)))
        indices = sorted(indices, key=functools.cmp_to_key(cmp))
        return self.new_mask_index(self, indices)

    def __getitem__(self, n):
        if isinstance(n, slice):
            return self.from_slice(n)
        if isinstance(n, (tuple, list, np.ndarray)):
            return self.from_multi(n)
        if isinstance(n, dict):
            return self.from_dict(n)
        return self._getitem(n)

    def from_slice(self, s):
        indices = range(len(self))[s]
        return self.new_mask_index(self, indices)

    def from_mask(self, lst):
        indices = [i for i, x in enumerate(lst) if x]
        return self.new_mask_index(self, indices)

    def from_multi(self, a):
        # will raise IndexError if an index is out of bounds
        n = len(self)
        indices = np.arange(0, n if n > 0 else 0)
        indices = indices[a].tolist()
        return self.new_mask_index(self, indices)

    def from_dict(self, dic):
        return self.sel(dic)

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, Index) for _ in sources)
        return MultiIndex(sources)

    def to_numpy(self, *args, **kwargs):
        import numpy as np

        return np.array([f.to_numpy(*args, **kwargs) for f in self])

    def full(self, *coords):
        return FullIndex(self, *coords)


class MaskIndex(Index):
    def __init__(self, index, indices):
        self.index = index
        self.indices = list(indices)
        # super().__init__(
        #     *self.index._init_args,
        #     order_by=self.index._init_order_by,
        #     **self.index._init_kwargs,
        # )

    def _getitem(self, n):
        n = self.indices[n]
        return self.index[n]

    def __len__(self):
        return len(self.indices)

    def __repr__(self):
        return "MaskIndex(%r,%s)" % (self.index, self.indices)


class MultiIndex(Index):
    def __init__(self, indexes, *args, **kwargs):
        self.indexes = list(indexes)
        super().__init__(*args, **kwargs)
        # self.indexes = list(i for i in indexes if len(i))
        # TODO: propagate  index._init_args, index._init_order_by, index._init_kwargs, for each i in indexes?

    def sel(self, *args, **kwargs):
        if not args and not kwargs:
            return self
        return self.__class__(i.sel(*args, **kwargs) for i in self.indexes)

    def _getitem(self, n):
        k = 0
        while n >= len(self.indexes[k]):
            n -= len(self.indexes[k])
            k += 1
        return self.indexes[k][n]

    def __len__(self):
        return sum(len(i) for i in self.indexes)

    def graph(self, depth=0):
        print(" " * depth, self.__class__.__name__)
        for s in self.indexes:
            s.graph(depth + 3)

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ",".join(repr(i) for i in self.indexes),
        )


class ForwardingIndex(Index):
    def __init__(self, index):
        self.index = index

    def __len__(self):
        return len(self.index)


class ScaledField:
    def __init__(self, field, offset, scaling):
        self.field = field
        self.offset = offset
        self.scaling = scaling

    def to_numpy(self, **kwargs):
        return (self.field.to_numpy(**kwargs) - self.offset) * self.scaling


class ScaledIndex(ForwardingIndex):
    def __init__(self, index, offset, scaling):
        super().__init__(index)
        self.offset = offset
        self.scaling = scaling

    def __getitem__(self, n):
        return ScaledField(self.index[n], self.offset, self.scaling)


class FullIndex(Index):
    def __init__(self, index, *coords):
        import numpy as np

        self.index = index

        # Pass1, unique values
        unique = index.unique_values(*coords)
        shape = tuple(len(v) for v in unique.values())

        name_to_index = defaultdict(dict)

        for k, v in unique.items():
            for i, e in enumerate(v):
                name_to_index[k][e] = i

        self.size = math.prod(shape)
        self.shape = shape
        self.holes = np.full(shape, False)

        for f in index:
            idx = tuple(name_to_index[k][f.metadata(k, default=None)] for k in coords)
            self.holes[idx] = True

        self.holes = self.holes.flatten()
        print("+++++++++", self.holes.shape, coords, self.shape)

    def __len__(self):
        return self.size

    def _getitem(self, n):
        assert self.holes[n], f"Attempting to access hole {n}"
        return self.index[sum(self.holes[:n])]



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
import os

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.testing import earthkit_examples_file

LOG = logging.getLogger(__name__)


def test_file_source_grib():
    s = from_source("file", earthkit_examples_file("test.grib"))
    from earthkit.data.readers.grib.reader import GRIBReader

    assert isinstance(s, GRIBReader), s
    assert len(s) == 2


def test_file_source_netcdf():
    s = from_source("file", earthkit_examples_file("test.nc"))
    assert len(s) == 2


def test_file_source_odb():
    s = from_source("file", earthkit_examples_file("test.odb"))
    assert s.path == earthkit_examples_file("test.odb")


# def test_user_1():
#     s = from_source("file", earthkit_file("docs/examples/test.grib"))
#     home_file = os.path.expanduser("~/.earthkit_data/test.grib")
#     try:
#         s.save(home_file)
#         # Test expand user
#         s = from_source("file", "~/.earthkit_data/test.grib")
#         assert len(s) == 2
#     finally:
#         try:
#             os.unlink(home_file)
#         except OSError:
#             LOG.exception("unlink(%s)", home_file)


# @pytest.mark.skipif(
#     sys.platform == "win32", reason="Cannot (yet) use expandvars on Windows"
# )
# def test_user_2():
#     s = from_source("file", earthkit_file("docs/examples/test.grib"))
#     home_file = os.path.expanduser("~/.earthkit_data/test.grib")
#     try:
#         s.save(home_file)
#         # Test expand vars
#         s = from_source("file", "$HOME/.earthkit_data/test.grib", expand_vars=True)
#         assert len(s) == 2
#     finally:
#         try:
#             os.unlink(home_file)
#         except OSError:
#             LOG.exception("unlink(%s)", home_file)


def test_glob():
    s = from_source("file", earthkit_examples_file("test.grib"))
    with temp_directory() as tmpdir:
        s.save(os.path.join(tmpdir, "a.grib"))
        s.save(os.path.join(tmpdir, "b.grib"))

        s = from_source("file", os.path.join(tmpdir, "*.grib"))
        assert len(s) == 4, len(s)

        s = from_source("file", tmpdir)
        assert len(s) == 4, len(s)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from earthkit.data import cache, from_source, settings
from earthkit.data.core.caching import cache_file
from earthkit.data.core.temporary import temp_directory
from earthkit.data.testing import earthkit_examples_file


def check_cache_files(dir_path):
    def touch(target, args):
        assert args["foo"] in (1, 2)
        with open(target, "w"):
            pass

        path1 = cache_file(
            "test_cache",
            touch,
            {"foo": 1},
            extension=".test",
        )

        path2 = cache_file(
            "test_cache",
            touch,
            {"foo": 2},
            extension=".test",
        )

        assert os.path.exists(path1)
        assert os.path.exists(path2)
        assert os.path.dirname(path1) == dir_path
        assert os.path.dirname(path1) == dir_path
        assert path1 != path2

        cnt = 0
        for f in cache.cache_entries():
            if f["owner"] == "test_cache":
                cnt += 1

        assert cnt == 2


def test_cache_1():
    with settings.temporary():
        settings.set("maximum-cache-disk-usage", "99%")
        cache.purge_cache(matcher=lambda e: ["owner"] == "test_cache")
        check_cache_files(settings.get("user-cache-directory"))


# 1GB ram disk on MacOS (blocks of 512 bytes)
# diskutil erasevolume HFS+ "RAMDisk" `hdiutil attach -nomount ram://2097152`
@pytest.mark.skipif(not os.path.exists("/Volumes/RAMDisk"), reason="No RAM disk")
def test_cache_4():
    with settings.temporary():
        settings.set("cache-directory", "/Volumes/RAMDisk/earthkit_data")
        settings.set("maximum-cache-disk-usage", "90%")
        for n in range(10):
            from_source("dummy-source", "zeros", size=100 * 1024 * 1024, n=n)


def test_cache_policy():
    with temp_directory() as user_dir:
        # cache = user dir
        with settings.temporary():
            settings.set({"cache-policy": "user", "user-cache-directory": user_dir})
            assert settings.get("cache-policy") == "user"
            assert settings.get("user-cache-directory") == user_dir
            assert cache.policy.has_cache() is True
            cache_dir = cache.policy.cache_directory()
            assert cache_dir == user_dir
            assert os.path.exists(cache_dir)
            check_cache_files(cache_dir)

            # cache = temporary with auto generated path
            with settings.temporary(
                {"cache-policy": "temporary", "temporary-cache-directory-root": None}
            ):
                assert settings.get("cache-policy") == "temporary"
                assert settings.get("temporary-cache-directory-root") is None
                assert cache.policy.has_cache() is True
                cache_dir = cache.policy.cache_directory()
                assert os.path.exists(cache_dir)
                check_cache_files(cache_dir)

            # cache = user dir (again)
            assert settings.get("cache-policy") == "user"
            assert settings.get("user-cache-directory") == user_dir
            assert cache.policy.has_cache() is True
            cache_dir = cache.policy.cache_directory()
            assert cache_dir == user_dir
            assert os.path.exists(cache_dir)
            check_cache_files(cache_dir)

            # cache = temporary with user defined root path
            with temp_directory() as root_dir:
                with settings.temporary(
                    {
                        "cache-policy": "temporary",
                        "temporary-cache-directory-root": root_dir,
                    }
                ):
                    assert settings.get("cache-policy") == "temporary"
                    assert settings.get("temporary-cache-directory-root") == root_dir
                    assert cache.policy.has_cache() is True
                    cache_dir = cache.policy.cache_directory()
                    assert os.path.exists(cache_dir)
                    os.path.dirname(cache_dir) == root_dir
                    check_cache_files(cache_dir)

            # cache = off
            with settings.temporary("cache-policy", "off"):
                assert settings.get("cache-policy") == "off"
                assert settings.get("user-cache-directory") == user_dir
                assert cache.policy.has_cache() is False
                assert cache.policy.cache_directory() is None

                with pytest.raises(RuntimeError):
                    cache_file(
                        "dummy_test_cache",
                        None,
                        {"foo": 1},
                        extension=".test",
                    )

            # cache = user dir (again)
            assert settings.get("cache-policy") == "user"
            assert settings.get("user-cache-directory") == user_dir
            assert cache.policy.has_cache() is True
            cache_dir = cache.policy.cache_directory()
            assert cache_dir == user_dir
            assert os.path.exists(cache_dir)
            check_cache_files(cache_dir)


def test_url_source_no_cache():
    with settings.temporary("cache-policy", "off"):
        with pytest.raises(RuntimeError):
            from_source(
                "url",
                "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test.grib",
            )


def test_grib_no_cache():
    with settings.temporary("cache-policy", "off"):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        f = ds[3]
        assert f.metadata("param") == "t"


@pytest.mark.parametrize("index_cache", [True, False])
def test_grib_offset_index_cache(index_cache):
    s = {"cache-policy": "temporary", "use-message-position-index-cache": index_cache}
    with settings.temporary(s):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        f = ds[3]
        assert f.metadata("param") == "t", f"index-cache={index_cache}"


# See github #155. This test can hang so we must set a timeout.
@pytest.mark.no_cache_init
@pytest.mark.timeout(20)
def test_cache_with_log_debug(caplog):
    import logging

    # the cache must not be initialised at this point
    assert cache._policy is None
    assert cache._manager is None

    caplog.set_level(logging.DEBUG)
    LOG = logging.getLogger(__name__)

    class A:
        def __repr__(self):
            d = cache.cache_directory()
            return d

    a = A()
    LOG.debug(f"dir {a}")
    # NOTE: if we use "%s" formatting e.g. "LOG.debug("dir %s", a)"
    # the problem still occurs!


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from . import Writer


class GribWriter(Writer):
    METADATA_TYPE = "GribMetadata"

    def write(self, f, values, metadata, check_nans=False):
        handle = metadata._handle
        if check_nans:
            import numpy as np

            if np.isnan(values).any():
                # missing_value = np.finfo(values.dtype).max
                missing_value = 9999
                values = np.nan_to_num(values, nan=missing_value)
                handle.set_double("missingValue", missing_value)
                handle.set_long("bitmapPresent", 1)

        handle.set_values(values)
        handle.write(f)


Writer = GribWriter



#!/usr/bin/env python3

# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data.decorators import normalize
from earthkit.data.vocabularies.aliases import unalias


def test_unalias():
    assert unalias("grib-paramid", "167") == "2t"
    assert unalias("grib-paramid", "2t") == "2t"


def func_x(x):
    return x


def test_aliases_grib_paramid_mutiple_false():
    aliases_grib_paramid = normalize(
        "x",
        type=str,
        aliases="grib-paramid",
        multiple=False,
    )(func_x)
    assert aliases_grib_paramid("u") == "u"
    assert aliases_grib_paramid(131) == "u"
    assert aliases_grib_paramid("131") == "u"

    # one-element list/tuple
    assert aliases_grib_paramid(("131",)) == "u"
    assert aliases_grib_paramid(["131"]) == "u"

    # list/tuple
    with pytest.raises(TypeError):
        aliases_grib_paramid(["131", "v"])

    # empty list/tuple
    with pytest.raises(TypeError):
        aliases_grib_paramid([])
    with pytest.raises(TypeError):
        aliases_grib_paramid(tuple([]))


def test_aliases_grib_paramid_mutiple_true():
    aliases_grib_paramid = normalize(
        "x",
        type=str,
        aliases="grib-paramid",
        multiple=True,
    )(func_x)
    # single values
    assert aliases_grib_paramid("u") == ["u"]
    assert aliases_grib_paramid(131) == ["u"]
    assert aliases_grib_paramid("131") == ["u"]

    # one-element list/tuple
    assert aliases_grib_paramid(("131",)) == ["u"]
    assert aliases_grib_paramid(["131"]) == ["u"]

    # list/tuple
    assert aliases_grib_paramid(["131", "v"]) == ["u", "v"]
    assert aliases_grib_paramid([131, "v"]) == ["u", "v"]
    assert aliases_grib_paramid(["u", "v"]) == ["u", "v"]
    assert aliases_grib_paramid(("u", "v")) == ["u", "v"]
    assert aliases_grib_paramid([]) == []
    assert aliases_grib_paramid(tuple([])) == []


def test_aliases_grib_paramid_mutiple_none():
    aliases_grib_paramid = normalize(
        "x",
        type=str,
        aliases="grib-paramid",
    )(func_x)
    # single values
    assert aliases_grib_paramid(131) == "u"
    assert aliases_grib_paramid("u") == "u"
    assert aliases_grib_paramid("131") == "u"

    # one-element list/tuple
    assert aliases_grib_paramid(("131",)) == ("u",)
    assert aliases_grib_paramid(["131"]) == ["u"]

    # list/tuple
    assert aliases_grib_paramid(["131", "v"]) == ["u", "v"]
    assert aliases_grib_paramid([131, "v"]) == ["u", "v"]
    assert aliases_grib_paramid(["u", "v"]) == ["u", "v"]
    assert aliases_grib_paramid(("u", "v")) == ("u", "v")
    assert aliases_grib_paramid([]) == []
    assert aliases_grib_paramid(tuple([])) == ()


if __name__ == "__main__":
    # import logging

    # logging.basicConfig(level=logging.WARNING)
    # test_aliases_grib_paramid_mutiple_true()
    # test_aliases_grib_paramid_mutiple_none()
    from earthkit.data.testing import main

    main(__file__)
